#!/usr/bin/env python3

# Copyright (c) 2019-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""
This script runs an SO time domain simulation.

You can see the automatically generated command line options with:

    toast_so_sim --help

Or you can dump a config file with all the default values with:

    toast_so_sim --default_toml config.toml

This script contains just comments about what is going on.  For details about all the
options for a specific Operator, see the documentation or use the help() function from
an interactive python session.

"""

import argparse
import datetime
import os
import sys
import traceback

import numpy as np

from astropy import units as u

# Import sotodlib.toast first, since that sets default object names
# to use in toast.
import sotodlib.toast as sotoast

import toast
import toast.ops

from toast.mpi import MPI, Comm
from toast.observation import default_values as defaults

from .. import ops as so_ops
from .. import workflows as wrk

# Make sure pixell uses a reliable FFT engine
import pixell.fft

pixell.fft.engine = "fftw"


def simulate_data(job, otherargs, runargs, data):
    log = toast.utils.Logger.get()

    wrk.simulate_atmosphere_signal(job, otherargs, runargs, data)

    # Shortcut if we are only caching the atmosphere.  If this job is only caching
    # (not observing) the atmosphere, then return at this point.
    if job.operators.sim_atmosphere.cache_only:
        return

    wrk.simulate_sky_map_signal(job, otherargs, runargs, data)
    wrk.simulate_conviqt_signal(job, otherargs, runargs, data)
    wrk.simulate_scan_synchronous_signal(job, otherargs, runargs, data)
    wrk.simulate_source_signal(job, otherargs, runargs, data)
    wrk.simulate_sso_signal(job, otherargs, runargs, data)
    wrk.simulate_catalog_signal(job, otherargs, runargs, data)
    wrk.simulate_wiregrid_signal(job, otherargs, runargs, data)
    wrk.simulate_stimulator_signal(job, otherargs, runargs, data)
    wrk.simulate_detector_timeconstant(job, otherargs, runargs, data)
    wrk.simulate_mumux_crosstalk(job, otherargs, runargs, data)
    wrk.simulate_detector_noise(job, otherargs, runargs, data)
    wrk.simulate_hwpss_signal(job, otherargs, runargs, data)
    wrk.simulate_detector_yield(job, otherargs, runargs, data)
    wrk.simulate_calibration_error(job, otherargs, runargs, data)
    wrk.simulate_readout_effects(job, otherargs, runargs, data)

    comm = data.comm.comm_world

    mem = toast.utils.memreport(msg="(whole node)", comm=comm, silent=True)
    log.info_rank(f"After simulating data:  {mem}", comm)

    wrk.save_data_hdf5(job, otherargs, runargs, data)

    mem = toast.utils.memreport(msg="(whole node)", comm=comm, silent=True)
    log.info_rank(f"After saving data:  {mem}", comm)


def reduce_data(job, otherargs, runargs, data):
    log = toast.utils.Logger.get()

    wrk.preprocess(job, otherargs, runargs, data)

    wrk.raw_statistics(job, otherargs, runargs, data)

    wrk.diff_noise_estimation(job, otherargs, runargs, data)
    wrk.noise_estimation(job, otherargs, runargs, data)

    data = wrk.demodulate(job, otherargs, runargs, data)

    wrk.filter_ground(job, otherargs, runargs, data)
    wrk.filter_common_mode(job, otherargs, runargs, data)
    wrk.filter_poly1d(job, otherargs, runargs, data)
    wrk.filter_poly2d(job, otherargs, runargs, data)

    wrk.diff_noise_estimation(job, otherargs, runargs, data)
    wrk.noise_estimation(job, otherargs, runargs, data)

    wrk.processing_mask(job, otherargs, runargs, data)
    wrk.flag_sso(job, otherargs, runargs, data)
    wrk.hn_map(job, otherargs, runargs, data)
    wrk.cadence_map(job, otherargs, runargs, data)
    wrk.crosslinking_map(job, otherargs, runargs, data)

    if job.operators.splits.enabled:
        wrk.splits(job, otherargs, runargs, data)
    else:
        wrk.mapmaker_ml(job, otherargs, runargs, data)
        wrk.mapmaker(job, otherargs, runargs, data)
        wrk.mapmaker_filterbin(job, otherargs, runargs, data)
        wrk.mapmaker_madam(job, otherargs, runargs, data)
    wrk.filtered_statistics(job, otherargs, runargs, data)

    mem = toast.utils.memreport(
        msg="(whole node)", comm=data.comm.comm_world, silent=True
    )
    log.info_rank(f"After reducing data:  {mem}", data.comm.comm_world)


def main():
    env = toast.utils.Environment.get()
    log = toast.utils.Logger.get()
    gt = toast.timing.GlobalTimers.get()
    gt.start("toast_so_sim (total)")
    timer = toast.timing.Timer()
    timer.start()

    # Get optional MPI parameters
    comm, procs, rank = toast.get_world()

    # If the user has not told us to use multiple threads,
    # then just use one.

    if "OMP_NUM_THREADS" in os.environ:
        nthread = os.environ["OMP_NUM_THREADS"]
    else:
        nthread = 1
    log.info_rank(
        f"Executing workflow with {procs} MPI tasks, each with "
        f"{nthread} OpenMP threads at {datetime.datetime.now()}",
        comm,
    )

    mem = toast.utils.memreport(msg="(whole node)", comm=comm, silent=True)
    log.info_rank(f"Start of the workflow:  {mem}", comm)

    # Argument parsing
    parser = argparse.ArgumentParser(description="SO simulation pipeline")

    parser.add_argument(
        "--out_dir",
        required=False,
        type=str,
        default="toast_out",
        help="The output directory",
    )
    parser.add_argument(
        "--obsmaps",
        required=False,
        default=False,
        action="store_true",
        help="Map each observation separately.",
    )
    parser.add_argument(
        "--detmaps",
        required=False,
        default=False,
        action="store_true",
        help="Map each detector separately.",
    )
    parser.add_argument(
        "--intervalmaps",
        required=False,
        default=False,
        action="store_true",
        help="Map each interval separately.",
    )
    parser.add_argument(
        "--zero_loaded_data",
        required=False,
        default=False,
        action="store_true",
        help="Zero out detector data loaded from disk",
    )
    parser.add_argument(
        "--preprocess_copy",
        required=False,
        default=False,
        action="store_true",
        help="Perform preprocessing on a copy of the data.",
    )
    parser.add_argument(
        "--log_config",
        required=False,
        default=None,
        help="Dump out config log to this yaml file",
    )

    # The operators and templates we want to configure from the command line
    # or a parameter file.

    operators = list()
    templates = list()

    wrk.setup_load_or_simulate_observing(parser, operators)
    wrk.setup_preprocess(parser, operators)

    wrk.setup_simulate_atmosphere_signal(operators)
    wrk.setup_simulate_sky_map_signal(operators)
    wrk.setup_simulate_conviqt_signal(operators)
    wrk.setup_simulate_scan_synchronous_signal(operators)
    wrk.setup_simulate_source_signal(operators)
    wrk.setup_simulate_sso_signal(operators)
    wrk.setup_simulate_catalog_signal(operators)
    wrk.setup_simulate_wiregrid_signal(operators)
    wrk.setup_simulate_stimulator_signal(operators)
    wrk.setup_simulate_detector_timeconstant(operators)
    wrk.setup_simulate_mumux_crosstalk(operators)
    wrk.setup_simulate_detector_noise(operators)
    wrk.setup_simulate_hwpss_signal(operators)
    wrk.setup_simulate_detector_yield(operators)
    wrk.setup_simulate_calibration_error(operators)
    wrk.setup_simulate_readout_effects(operators)
    wrk.setup_save_data_hdf5(operators)

    wrk.setup_raw_statistics(operators)

    wrk.setup_filter_common_mode(operators)
    wrk.setup_filter_ground(operators)
    wrk.setup_filter_poly1d(operators)
    wrk.setup_filter_poly2d(operators)
    wrk.setup_noise_estimation(operators)

    wrk.setup_demodulate(operators)

    wrk.setup_processing_mask(operators)
    wrk.setup_flag_sso(operators)
    wrk.setup_hn_map(operators)
    wrk.setup_cadence_map(operators)
    wrk.setup_crosslinking_map(operators)

    wrk.setup_mapmaker_ml(operators)
    wrk.setup_mapmaker(operators, templates)
    wrk.setup_mapmaker_filterbin(operators)
    wrk.setup_mapmaker_madam(operators)
    wrk.setup_splits(operators)
    wrk.setup_filtered_statistics(operators)

    job, config, otherargs, runargs = wrk.setup_job(
        parser=parser, operators=operators, templates=templates
    )

    # Create our output directory
    if comm is None or comm.rank == 0:
        if not os.path.isdir(otherargs.out_dir):
            os.makedirs(otherargs.out_dir, exist_ok=True)

    # Log the config that was actually used at runtime.
    if otherargs.log_config is not None:
        # User wants to specify the location of the yaml config log
        toast.config.dump_yaml(otherargs.log_config, config, comm=comm)
    else:
        # Log a toml file in the output directory
        outlog = os.path.join(otherargs.out_dir, "config_log.toml")
        toast.config.dump_toml(outlog, config, comm=comm)

    # If this is a dry run, exit
    if otherargs.dry_run:
        log.info_rank("Dry-run complete", comm=comm)
        return

    data = wrk.load_or_simulate_observing(job, otherargs, runargs, comm)

    simulate_data(job, otherargs, runargs, data)

    if not job.operators.sim_atmosphere.cache_only:
        # Reduce the data
        reduce_data(job, otherargs, runargs, data)

    # Collect optional timing information
    alltimers = toast.timing.gather_timers(comm=comm)
    if data.comm.world_rank == 0:
        out = os.path.join(otherargs.out_dir, "timing")
        toast.timing.dump(alltimers, out)

    log.info_rank("Workflow completed in", comm=comm, timer=timer)


def cli():
    world, procs, rank = toast.mpi.get_world()
    with toast.mpi.exception_guard(comm=world):
        main()


if __name__ == "__main__":
    cli()



# Copyright (c) 2020-2025 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simons Observatory processing modules.

"""

# Namespace imports

from .act_sign import ActSign
from .bias_cut import DetBiasCuts
from .corotator import CoRotator
from .sim_sso import SimSSO
from .sim_catalog import SimCatalog
from .sim_hwpss import SimHWPSS
from .sim_source import SimSource
from .h_n import Hn
from .mlmapmaker import MLMapmaker
from .sim_wiregrid import SimWireGrid
from .sim_stimulator import SimStimulator
from .save_books import SaveBooks
from .load_books import LoadBooks
from .sim_readout import SimReadout
from .sim_mumux_crosstalk import SimMuMUXCrosstalk
from .splits import Splits
from .mumux_crosstalk_util import detmap_available, pos_to_chi
from .load_context import LoadContext
from .readout_filter import ReadoutFilter
from .dark_template import DarkTemplate
from .trend_cut import DetTrendCuts



import numpy as np
import pyfftw
import inspect
import scipy.signal as signal
from operator import attrgetter

import logging

from . import detrend_tod
from . import fft_ops
from sotodlib import core

logger = logging.getLogger(__name__)


def fourier_filter(tod, filt_function,
                   detrend=None, resize='zero_pad',
                   axis_name='samps', signal_name='signal', 
                   time_name='timestamps', rfft=None,
                   **kwargs):
    """Return a filtered tod.signal_name along the axis axis_name. 
        Does not change the data in the axis manager.
    
    Arguments:
    
        tod: axis manager
        
        filt_function: function( freqs, tod ) function that takes a set of 
            frequencies and the axis manager and returns the filter in 
            fourier space
        
        detrend: Method of detrending to be done before ffting. Can
            be 'linear', 'mean', or None. Note that detrending here can be slow
            for large arrays
            
        resize: How to resize the axis to increase fft
            speed. 'zero_pad' will increase to the next nice number (a
            product of small primes compatible with the FFT
            implementation).  'trim' will eliminate samples from the
            end so that axis has a nice length for FFTs.  None will not
            change the axis length and might be quite slow. Trim will
            be kinda weird here, because signal will not be returned
            as the same size as it is input

        axis_name: name of axis you would like to fft along
        
        signal_name: name of the variable in tod to fft
        
        time_name: name for getting time of data (in seconds) from tod

        rfft: A RFFTObj for performing the FFTs.
              Must be valid for the input data and contain both t_forward and t_backward.
        
    Returns:
    
        signal: filtered tod.signal_name 
        
    """
    axis = getattr(tod, axis_name)
    times = getattr(tod, time_name)
    delta_t = (times[-1]-times[0])/axis.count
    
    if len(tod._assignments[signal_name])==1:
        n_det = 1
        ## signal will be at least 2D
        other_idx = None
    elif len(tod._assignments[signal_name])==2:
        checks = np.array([x==axis_name for x in tod._assignments[signal_name]],dtype='bool')
        other_idx = np.where(~checks)[0][0]
        other_axis = getattr(tod, tod._assignments[signal_name][other_idx])
        n_det = other_axis.count
    else:
        raise ValueError('fourier_filter only works for 1D or 2D data streams')
    
    if resize == 'zero_pad':
        n = fft_ops.find_superior_integer(axis.count)
        logger.info('fourier_filter: padding %i -> %i' % (axis.count, n))
    elif resize == 'trim':
        n = fft_ops.find_inferior_integer(axis.count)
        logger.info('fourier_filter: trimming %i -> %i' % (axis.count, n))
    elif resize is None:
        n = axis.count
    else:
        raise ValueError('resize must be "zero_pad", "trim", or None')

    if detrend is not None:
        logger.info('fourier_filter: detrending.')
        signal = detrend_tod(tod, detrend, axis_name=axis_name,
                             signal_name=signal_name, in_place=False)
    else:
        signal = tod[signal_name]
    signal = np.atleast_2d(signal)

    if (signal.dtype != "float32"):
        logger.warn(f'fourier_filter: signal dtype {signal.dtype} is not float32. '
                    f'fourier_filter will demote to float32.')

    if isinstance(filt_function, identity_filter):
        logger.info('fourier_filter: filt_function is identity; skipping FFT.')
        signal = signal.copy()

    else:
        if rfft is None:
            logger.info('fourier_filter: initializing rfft object.')
            rfft = fft_ops.RFFTObj.for_shape(n_det, n, 'BOTH')
        else:
            logger.info('fourier_filter: using provided rfft object.')
            rfft.validate((n_det, n))

        if other_idx is not None and other_idx != 0:
            ## so that code can be written always along axis 1
            signal = signal.transpose()

        # This copy is valid for all modes of "resize"
        logger.info('fourier_filter: copying in data.')
        rfft.a[:,:min(n, axis.count)] = signal[:,:min(n, axis.count)]
        rfft.a[:,min(n, axis.count):] = 0

        ## FFT Signal
        logger.info('fourier_filter: FFT.')
        rfft.t_forward()

        ## Get Filter
        logger.info('fourier_filter: applying filter.')
        freqs = np.fft.rfftfreq(n, delta_t)
        filt_function.apply(freqs, tod, rfft.b, **kwargs)

        ## FFT Back
        logger.info('fourier_filter: IFFT.')
        rfft.t_backward()

        # Un-pad?
        signal = rfft.a[:,:min(n, axis.count)]

        if other_idx is not None and other_idx != 0:
            return signal.transpose()

    if other_idx is None:
        return signal[0]
    
    return signal

def fft_trim(tod, axis='samps', prefer='right'):
    """Restrict AxisManager sample range so that FFTs are efficient.  This
    uses the find_inferior_integer function.

    Args:
      tod (AxisManager): Target, which is modified in place.
      axis (str): Axis to target.
      prefer (str): One of ['left', 'right', 'center'], indicating
        whether to trim away samples from the end, the beginning, or
        !equally at the beginning and end (respectively).

    Returns:
      The (start, stop) indices to use to slice an array and get these
      samples.

    """
    axis_obj = tod[axis]
    old_size = axis_obj.count
    new_size = fft_ops.find_inferior_integer(old_size)

    offset = old_size - new_size
    if prefer == 'left':
        offset = 0
    elif prefer == 'center':
        offset //= 2
    elif prefer == 'right':
        pass
    else:
        raise ValueError(f'Invalid choice prefer="{prefer}"')

    start_stop = (offset, offset+new_size)

    if isinstance(axis_obj, core.OffsetAxis):
        # Account for special indexing of OffsetAxis.
        offset += axis_obj.offset

    tod.restrict(axis, (offset, offset+new_size))
    return start_stop


################################################################

# Base class... provides that a * b always returns a FilterChain.
class _chainable:
    @staticmethod
    def _preference(f):
        return getattr(f, 'preference', 'compose')
    def __mul__(self, other):
        if isinstance(other, identity_filter):
            return self
        if isinstance(self, identity_filter):
            return other
        return FilterChain([self, other])

class FilterFunc(_chainable):
    """Class to support chaining of Fourier filters.

    FilterFunc.deco may be used to decorate functions with signatures
    like::

       function_name(freqs, tod, *args, **kwargs)

    """
    # Note self._fun is set in the subclass, as a class variable,
    # e.g.: _fun = staticmethod(gaussian_filter)
    _fun_nargs = 2
    preference = 'compose'
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.args = args
        self.kwargs = kwargs
    def __call__(self, freqs, tod):
        return self._fun(freqs, tod, *self.args, **self.kwargs)
    def apply(self, freqs, tod, target):
        target *= self(freqs, tod)
    @classmethod
    def deco(cls, fun):
        class filter_func(cls):
            _fun = staticmethod(fun)
        # get help from someone
        filter_func.__doc__ = fun.__doc__
        # get arguments from someone after removing the partial args
        args = list(inspect.signature(fun).parameters.values())[cls._fun_nargs:]
        filter_func.__signature__ = inspect.Signature(parameters=args)
        return filter_func

class FilterApplyFunc(FilterFunc):
    """Class to support chaining of Fourier filters.

    @FilterApplyFunc.deco may be used to decorate functions with signatures
    like:

       function_name(target, freqs, tod, *, **)

    Such filter functions must return a transfer function array, if
    target=None.  But if target is a fourier transform array, the
    function must apply the transfer function to that array and return
    None.

    """
    _fun_nargs = 3
    preference = 'apply'
    def __call__(self, freqs, tod):
        return self._fun(None, freqs, tod, *self.args, **self.kwargs)
    def apply(self, freqs, tod, target):
        return self._fun(target, freqs, tod, *self.args, **self.kwargs)

class FilterChain(_chainable):
    """A chain of Fourier filters."""
    def __init__(self, items):
        super().__init__()
        self.links = []
        for a in items:
            if isinstance(a, FilterChain):
                self.links.extend(a.links)
            else:
                self.links.append(a)

    def __call__(self, freqs, tod):
        # We could do better by handling self.links out of order.
        filt = self.links[0](freqs, tod)
        for f in self.links[1:]:
            if self._preference(f) == 'apply' and filt.ndim == 2:
                f.apply(freqs, tod, filt)
            else:
                _filt = f(freqs, tod)
                # Swap those to help broadcasting work...
                if _filt.ndim > filt.ndim:
                    filt, _filt = _filt, filt
                filt *= _filt
                del _filt
        return filt

    def apply(self, freqs, tod, target):
        filt = None
        for f in self.links:
            if self._preference(f) == 'apply':
                f.apply(freqs, tod, target)
            elif filt is None:
                filt = f(freqs, tod)
            else:
                _filt = f(freqs, tod)
                # Swap those to help broadcasting work...
                if _filt.ndim > filt.ndim:
                    filt, _filt = _filt, filt
                filt *= _filt
                del _filt
        if filt is not None:
            target *= filt

# Alias the decorators...
fft_filter = FilterFunc.deco
fft_apply_filter = FilterApplyFunc.deco


# Filtering Functions
#################
@fft_filter
def counter_1_over_f(freqs, tod, fk, n):
    """
    Counter 1/f filter for noise w/ PSD that follows:
    
    w*(1 + (fk/f)**n) 
    where w is the white noise level, fk is the knee frequency, and
    n is the 1/f index.
    """
    return 1/(1+(fk/freqs)**n)

@fft_filter
def identity_filter(freqs, tod, invert=False):
    """Identity filter (gain=1 at all frequencies).

    This filter has some special handling -- the FFT will not take
    place if your only intention is to apply an identify filter; also
    identity_filter() * other_filter() will simply return the
    other_filter().

    """
    return np.ones(len(freqs))

@fft_filter
def gain(freqs, tod, gain=1.):
    """Filter that simply applies a simple gain (which could be complex)
    to the entire spectrum.

    This can be used (with gain=1) as an identity filter for testing
    the filter preprocessing.

    """
    return gain * np.ones(len(freqs))

@fft_filter
def low_pass_butter4(freqs, tod, fc):
    """4th-order low-pass filter with f3db at fc (Hz).

    """
    b, a = signal.butter(4, 2*np.pi*fc, 'lowpass', analog=True)
    return np.abs(signal.freqs(b, a, 2*np.pi*freqs)[1])

@fft_filter
def high_pass_butter4(freqs, tod, fc):
    """4th-order high-pass filter with f3db at fc (Hz).

    """
    b, a = signal.butter(4, 2*np.pi*fc, 'highpass', analog=True)
    return np.abs(signal.freqs(b, a, 2*np.pi*freqs)[1])

@fft_filter
def band_pass_butter4(freqs, tod, fc_low, fc_high):
    """4th-order band-pass filter with f3db at fc (Hz).

    """
    b, a = signal.butter(4, [2*np.pi*fc_low, 2*np.pi*fc_high],
                         'bandpass', analog=True)
    return np.abs(signal.freqs(b, a, 2*np.pi*freqs)[1])

@fft_filter
def band_stop_butter4(freqs, tod, fc_low, fc_high):
    """4th-order band-stop filter with f3db at fc (Hz).

    """
    b, a = signal.butter(4, [2*np.pi*fc_low, 2*np.pi*fc_high],
                         'bandstop', analog=True)
    return np.abs(signal.freqs(b, a, 2*np.pi*freqs)[1])

@fft_filter
def tau_filter(freqs, tod, tau_name='timeconst', do_inverse=True):
    """tau_filter is deprecated; use timeconst_filter."""
    logging.warning('tau_filter is deprecated; use timeconst_filter.')
    taus = getattr(tod, tau_name)
    
    filt = 1 + 2.0j*np.pi*taus[:,None]*freqs[None,:]
    if not do_inverse:
        return 1.0/filt
    return filt

@fft_apply_filter
def timeconst_filter(target, freqs, tod, timeconst=None, invert=False):
    """One-pole time constant filter for fourier_filter.

    Builds filter for applying or removing time constants from signal
    data.

    Args:

      timeconst: Array of time constant values (one per detector).
        Alternately, a string indicating what member of tod to use for
        the time constants array.  Defaults to 'timeconst'.
      invert (bool): If true, returns the inverse transfer function,
        to deconvolve the time constants.

    Example::

      # Deconvolve time constants.
      fourier_filter(tod, timeconst_filter(invert=True),
                     detrend='linear', resize='zero_pad')

    """
    if timeconst is None:
        timeconst = 'timeconst'
    if isinstance(timeconst, str):
        # attrgetter used to retrieve a field multiple layers deep.
        _f = attrgetter(timeconst)
        timeconst = _f(tod)

    if target is None:
        filt = 1 + 2.0j*np.pi*timeconst[:,None]*freqs[None,:]
        if invert:
            return filt
        return 1.0/filt

    # Apply filter directly to FFT in target.
    assert(len(timeconst) == len(target))  # safe zip.
    if invert:
        for tau, dest in zip(timeconst, target):
            dest *= 1.+ 2.j*np.pi*tau*freqs
    else:
        for tau, dest in zip(timeconst, target):
            dest /= 1.+ 2.j*np.pi*tau*freqs

@fft_filter
def timeconst_filter_single(freqs, tod, timeconst, invert=False):
    """One-pole time constant filter for fourier_filter.

    This version accepts a single time constant value, in seconds.  To
    use different time constants for each detector, see
    timeconst_filter.

    Example::

      # Apply a 1ms time constant.
      fourier_filter(tod, timeconst_filter_single(timeconst=0.001),
                     detrend=None, resize='zero_pad')

    """
    if invert:
        return 1. + 2.j * np.pi * timeconst * freqs
    return 1. / (1. + 2.j * np.pi * timeconst * freqs)

@fft_filter
def gaussian_filter(freqs, tod, fc=0., f_sigma=None, gain=1.0, t_sigma=None):
    """Gaussian bandpass filter

    Parameters:
        fc (float0): Central frequency of the filter (peak of
            passband), in Hz.
        f_sigma (float): Standard deviation of the filter kernel, in
            Hz.
        gain (float): Gain of the filter.
        t_sigma (float): Instead of f_sigma, set t_sigma and f_sigma =
            1/(2 pi t_sigma) will be used.

    The filter kernel has the shape of a normal distribution, centered
    on fc with standard deviation f_sigma, and peak height gain.

    """
    if t_sigma is not None and f_sigma is not None:
        raise ValueError("User must not specify both f_sigma and t_sigma.")
    if t_sigma is not None:
        f_sigma = 1.0 / (2*np.pi*t_sigma)
    if f_sigma is None:
        raise ValueError('User must specify either f_sigma or t_sigma.')
    return gain * np.exp(-0.5*(np.abs(freqs)-fc)**2/f_sigma**2)

@fft_filter
def low_pass_sine2(freqs, tod, cutoff, width=None):
    """Low-pass filter.  Response falls from 1 to 0 between frequencies
    (cutoff - width/2, cutoff + width/2), with a sine-squared shape.

    """
    if width is None:
        width = cutoff * 2
    phase = np.pi * np.clip((abs(freqs) - cutoff) / width, -0.5, 0.5)
    return 0.5 - 0.5 * np.sin(phase)

@fft_filter
def high_pass_sine2(freqs, tod, cutoff, width=None):
    """High-pass filter.  Response rises from 0 to 1 between frequencies
    (cutoff - width/2, cutoff + width/2), with a sine-squared shape.

    """
    if width is None:
        width = cutoff * 2
    phase = np.pi * np.clip((abs(freqs) - cutoff) / width, -0.5, 0.5)
    return 0.5 + 0.5 * np.sin(phase)

@fft_filter
def band_stop_sine2(freqs, tod, fc_low, fc_high, width=None):
    """Band-stop filter.  Response falls/rises from 1/0 to 0/1 between frequencies
    (fc_low/fc_high - width/2, fc_low/fc_high + width/2), with a sine-squared shape.

    """
    return low_pass_sine2._fun(freqs, tod, fc_low, width) + \
        high_pass_sine2._fun(freqs, tod, fc_high, width)

@fft_filter
def iir_filter(freqs, tod, b=None, a=None, fscale=1., iir_params=None,
               invert=False):
    """Infinite impulse response (IIR) filter.  This sort of filter is
    used in digital applications as a low-pass filter prior to
    decimation.  The Smurf and MCE readout filters can both be
    expressed in this form.

    Args:
      b: numerator polynomial filter coefficients (z^0,z^1, ...)
      a: denominator coefficients
      fscale: scalar used to compute z = exp(-2j*pi*freqs*fscale).
        In general this should be equal to 1/f_orig, where f_orig
        is the original sampling frequency (before downsampling).
      iir_params: Alternative way to specify b, a, and fscale (see
        notes).
      invert: If true, returns denom/num instead of num/denom.

    Notes:
      The `b` and `a` coefficients are as implemented in
      scipy.signal.freqs, scipy.signal.butter, etc.  The "angular
      frequencies", `w`, are computed as 2*pi*freqs*fscale.

      If the filter parameters (b, a, fscale) are not passed in
      explicitly, they will be extracted from an AxisManager based on
      the argument iir_params, which must be a dict or AxisManager
      with keys "b", "a", and "fscale", or an AxisManager including 
      the sub-iir_params of each stream_id. In the later case, if 
      the filter parameters of each stream_id is different,
      raises an error.
      
      But note that:

      - If iir_params is a string, tod[iir_params] is used (and must
        be an AxisManager or dict).
      - If iir_params is None, that's the same as passing
        iir_params='iir_params'.

    """
    if a is None:
        # Get params from TOD?
        if iir_params is None:
            iir_params = 'iir_params'
        if isinstance(iir_params, str):
            iir_params = tod[iir_params]
        if (isinstance(iir_params, core.AxisManager)
            and 'a' not in iir_params._fields):
            # Check iir_param's uniformity
            _a = None
            for _field, _sub_iir_params in iir_params._fields.items():
                if isinstance(_sub_iir_params, core.AxisManager) and 'a' in _sub_iir_params._fields:
                    sub_iir_params = _sub_iir_params
                    if _a is None:
                        _a, _b, _fscale = sub_iir_params['a'], sub_iir_params['b'], sub_iir_params['fscale']
                    else:
                        if np.any(np.hstack([sub_iir_params['a'] != _a,
                                             sub_iir_params['b'] != _b,
                                             sub_iir_params['fscale'] != _fscale,])):
                            raise ValueError('iir parameters are not uniform.')
            iir_params = sub_iir_params
            # check if iir_params from axis manager are None
            if iir_params['a'] is None or iir_params['b'] is None:
                raise ValueError('axis manager iir parameters are empty')
        try:
            a = iir_params['a']
            b = iir_params['b']
            fscale = iir_params['fscale']
        except Exception as e:
            raise ValueError("Failed to extract filter parameters from "
                             f"iir_params={iir_params}.")

    z = np.exp(-2j*np.pi*freqs * fscale)
    B, A = np.polyval(b[::-1], z), np.polyval(a[::-1], z)
    if invert:
        return A / B
    return B / A


# Functions to derive low/high/band pass filter from configuration
##################################################################
def get_lpf(cfg):
    """
    Returns a low-pass filter based on the configuration.

    Args:
        cfg (dict): A dictionary containing the low-pass filter configuration.
            It must have the following keys:
            - "type": A string specifying the type of low-pass filter. Supported values are "identity", "butter4" and "sine2".
            - "cutoff": A float specifying the cutoff frequency of the low-pass filter.
            - "trans_width": A float specifying the transition width of the low-pass filter (only for "sine2" type).

    Returns:
        filters.fourier_filter: the low-pass filter.
    """
    if cfg['type'] == 'identity':
        return identity_filter()
    elif cfg['type'] == 'butter4':
        cutoff = cfg['cutoff']
        return low_pass_butter4(fc=cutoff)
    elif cfg['type'] == 'sine2':
        cutoff = cfg['cutoff']
        trans_width = cfg['trans_width']
        return low_pass_sine2(cutoff=cutoff, width=trans_width)
    else:
        raise ValueError('Unsupported filter type. Supported filters are `identity`, `butter4` and `sine2`')


def get_hpf(cfg):
    """
    Returns a high-pass filter based on the configuration.

    Args:
        cfg (dict): A dictionary containing the high-pass filter configuration.
            It must have the following keys:
            - "type": A string specifying the type of high-pass filter. Supported values are "identity", "butter4" and "sine2".
            - "cutoff": A float specifying the cutoff frequency of the high-pass filter.
            - "trans_width": A float specifying the transition width of the high-pass filter (only for "sine2" type).

    Returns:
        filters.fourier_filter: the high-pass filter.
    """
    if cfg['type'] == 'identity':
        return identity_filter()
    elif cfg['type'] == 'butter4':
        cutoff = cfg['cutoff']
        return high_pass_butter4(fc=cutoff)
    elif cfg['type'] == 'sine2':
        cutoff = cfg['cutoff']
        trans_width = cfg['trans_width']
        return high_pass_sine2(cutoff=cutoff, width=trans_width)
    else:
        raise ValueError('Unsupported filter type. Supported filters are `identity`, `butter4` and `sine2`')
        

def get_bpf(cfg):
    """
    Returns a band-pass filter based on the configuration.

    Args:
        cfg (dict): A dictionary containing the band-pass filter configuration.
            It must have the following keys:
            - "type": A string specifying the type of band-pass filter. Supported values are "identity", "butter4" and "sine2".
            - "center": A float specifying the center frequency of the band-pass filter.
            - "width": A float specifying the width of the band-pass filter.
            - "trans_width": A float specifying the transition width of the band-pass filter (only for "sine2" type).

    Returns:
        filters.fourier_filter: the band-pass filter.
    """
    if cfg['type'] == 'identity':
        return identity_filter()
    elif cfg['type'] == 'butter4':
        center = cfg['center']
        width = cfg['width']
        return band_pass_butter4(fc_low=center - width/2., fc_high=center + width/2.)
    elif cfg['type'] == 'sine2':
        center = cfg['center']
        width = cfg['width']
        trans_width = cfg['trans_width']
        return low_pass_sine2(cutoff=center + width/2., width=trans_width)*\
                high_pass_sine2(cutoff=center - width/2., width=trans_width)
    else:
        raise ValueError('Unsupported filter type. Supported filters are `identity`, `butter4` and `sine2`')


def get_bsf(cfg):
    """
    Returns a band-stop filter based on the configuration.

    Args:
        cfg (dict): A dictionary containing the band-stop filter configuration.
            It must have the following keys:
            - "type": A string specifying the type of band-stop filter. Supported values are "identity", "butter4" and "sine2".
            - "center": A float specifying the center frequency of the band-stop filter.
            - "width": A float specifying the width of the band-stop filter.
            - "trans_width": A float specifying the transition width of the band-stop filter (only for "sine2" type).

    Returns:
        filters.fourier_filter: the band-stop filter.
    """
    if cfg['type'] == 'identity':
        return identity_filter()
    elif cfg['type'] == 'butter4':
        center = cfg['center']
        width = cfg['width']
        return band_stop_butter4(fc_low=center - width / 2., fc_high=center + width / 2.)
    elif cfg['type'] == 'sine2':
        center = cfg['center']
        width = cfg['width']
        trans_width = cfg['trans_width']
        return band_stop_sine2(fc_low=center - width / 2., fc_high=center + width / 2, width=trans_width)
    else:
        raise ValueError('Unsupported filter type. Supported filters are `identity`, `butter4` and `sine2`')



import unittest
import numpy as np
import matplotlib.pyplot as plt
from sotodlib import core, tod_ops, sim_flags

def quick_dets_axis(n_det):
    return core.LabelAxis('dets', ['det%04i' % i for i in range(n_det)])

def quick_focal_plane(dets, scale=0.5):
    n_det = dets.count
    nrow = int(np.ceil((n_det/2)**2))
    entries = []
    gamma = 0.
    for r in range(nrow):
        y = r * scale
        for c in range(nrow):
            x = c * scale
            entries.extend([(x, y, gamma),
                            (x, y, gamma + 90)])
            gamma += 15.
    entries = entries[(n_det - len(entries)) // 2:]
    entries = entries[:n_det]
    fp = core.AxisManager(dets)
    for k, v in zip(['xi', 'eta', 'gamma'], np.transpose(entries)):
        v = (v - v.mean()) * np.pi/180
        fp.wrap(k, v, [(0, 'dets')])
    return fp

def quick_scan(tod, n_samp):
    az_min = 100.
    az_max = 120.
    v_az = 2.
    
    dt = tod.timestamps[1] - tod.timestamps[0]
    samps_1scan = (az_max - az_min) / (v_az * dt) # one scan duration
    num_scans = int(n_samp / samps_1scan)
    turnaround_idxes = np.linspace(0, n_samp, num_scans, dtype=int)
    turnaround_template = np.zeros(turnaround_idxes.shape)
    turnaround_template[::2] = az_min
    turnaround_template[1::2] = az_max
    az = np.interp(np.arange(0, n_samp), turnaround_idxes, turnaround_template)
    bs = core.AxisManager(tod.samps)
    bs.wrap_new('az'  , shape=('samps', ))[:] = az * np.pi/180
    bs.wrap_new('el'  , shape=('samps', ))[:] = 50 * np.pi/180
    bs.wrap_new('roll', shape=('samps', ))[:] = 0. * az
    return bs

def gen_1_over_f(n_samp, dt, fknee):
    tmp = np.random.random(size=n_samp) * 2 - 1
    S = np.fft.rfft(tmp)

    fknee = 1
    freqs = np.fft.rfftfreq(n_samp, dt)
    freqs[0] = freqs[1]
    fil = 1 + (freqs / fknee) ** -2
    S = S * fil
    s = np.fft.irfft(S)
    s /= np.max(np.abs(s))
    return s

def quick_tod(n_det, n_samp, dt=.005):
    dets = quick_dets_axis(n_det)
    tod = core.AxisManager(
        quick_dets_axis(n_det),
        core.OffsetAxis('samps', n_samp))
    tod.wrap('focal_plane', quick_focal_plane(tod.dets))
    timestamps = np.arange(n_samp) * dt
    
    tod.wrap_new('timestamps', shape=('samps', ))[:] = 1800000000. + timestamps
    tod.wrap('boresight', quick_scan(tod, n_samp=n_samp))
    tod.wrap_new('signal', shape=('dets', 'samps'), dtype='float32')
    
    tod.wrap('flags', core.FlagManager(tod.dets, tod.samps))
    fknee = 0.5
    for di, det in enumerate(tod.dets.vals):
        tod.signal[di] = gen_1_over_f(n_samp,dt, fknee)
    params = {'n_glitches': 10, 'sig_n_glitch': 10, 'h_glitch': 10,
              'sig_h_glitch': 2}
    sim_flags.add_random_glitches(tod, params=params, signal='signal',
                                  flag='glitches', overwrite=False)
    return tod

class SubpolyfTest(unittest.TestCase):
    "Test the subscan polyfilter functions"
    def test_subpolyf(self):
        # prepare tod
        n_det = 10
        n_samp = 200 * 60 * 3
        tod = quick_tod(n_det, n_samp)
        tod.boresight.az[:200*10] = tod.boresight.az[200*10] #non scanning part at the start of obs
        _ = tod_ops.flags.get_turnaround_flags(tod, truncate=False)
        tod_ori = tod.copy()
        
        # 
        for method in ['polyfit', 'legendre']:
            tod = tod_ori.copy()
            _ = tod_ops.sub_polyf.subscan_polyfilter(tod, degree=5, signal_name="signal", exclude_turnarounds=False,
                       mask='glitches', exclusive=True, method=method, in_place=True)
            
            masked_signal = np.ma.masked_array(tod.signal, tod.flags.glitches.mask())
            masked_signal_ori = np.ma.masked_array(tod_ori.signal, tod_ori.flags.glitches.mask())
            std_suppression = np.mean(np.std(masked_signal, axis=1) / np.std(masked_signal_ori, axis=1))
            print(f'({method}) std is suppressed by factor of {std_suppression:.2e}')
            self.assertTrue(std_suppression < 0.1)
            
if __name__ == '__main__':
    unittest.main()
    


import argparse
import datetime as dt
import time
from typing import Optional
from sqlalchemy import not_

from sotodlib.site_pipeline.monitor import Monitor
from sotodlib.site_pipeline.util import init_logger

from sotodlib.io.imprinter import (
    Imprinter,
    Books,
    BOUND,
    UNBOUND,
    UPLOADED,
    FAILED,
    DONE,
)

logger = init_logger(__name__, "update_book_plan: ")

def main(
    config: str,
    min_ctime: Optional[float] = None,
    max_ctime: Optional[float] = None,
    stream_ids: Optional[str] = None,
    force_single_stream: bool = False,
    update_delay: float = 1,
    update_delay_timecodes: Optional[float] = 7,
    min_ctime_timecodes: Optional[float] = None,
    max_ctime_timecodes: Optional[float] = None,
    from_scratch: bool = False,
    use_monitor: bool = False,
    ):
    """
    Update the book plan database with new data from the g3tsmurf database.

    Parameters
    ----------
    config : str
        Path to config file for imprinter
    min_ctime : Optional[float], optional
        The minimum ctime to include in the book plan, by default None
    max_ctime : Optional[float], optional
        The maximum ctime to include in the book plan, by default None
    stream_ids : Optional[str], optional
        The stream ids to consider, list supplied as a comma separated string
        (e.g. "1,2,3"), by default None
    force_single_stream : bool, optional
        If True, tream multi-wafer data as if it were single wafer data, by default False
    update_delay : float, optional
        The range of time to search through g3tsmurf db for new data in units of
        days, by default 1
    update_delay_timecodes : float, optional
        The range of time to search through g3tsmurf db for new data in units of
        days for timecode books, by default 7
    min_ctime_timecodes : Optional[float], optional
        The minimum ctime to include in the book plan for timecode (hk, smurf, stray) books, by default None
    max_ctime_timecodes : Optional[float], optional
        The maximum ctime to include in the book planor timecode (hk, smurf, stray) books, by default None
    from_scratch : bool, optional
        If True, start to search from beginning of time, by default False
    use_monitor : bool
        if True, will send monitor information to influx, set to false by
        default so we can use identical config files for development
    """
    if stream_ids is not None:
        stream_ids = stream_ids.split(",")

    imprinter = Imprinter(
        config, 
        db_args={'connect_args': {'check_same_thread': False}},
        logger=logger,
        make_db=from_scratch,
    )
    
    # leaving min_ctime and max_ctime as None will go through all available 
    # data, so preferreably set them to a reasonable range based on update_delay
    if not from_scratch and min_ctime is None:
        min_ctime = dt.datetime.now() - dt.timedelta(days=update_delay)
    if isinstance(min_ctime, dt.datetime):
        min_ctime = min_ctime.timestamp()
    if isinstance(max_ctime, dt.datetime):
        max_ctime = max_ctime.timestamp()

    # obs and oper books
    logger.info("Registering obs/oper Books")
    imprinter.update_bookdb_from_g3tsmurf(
        min_ctime=min_ctime, max_ctime=max_ctime,
        ignore_singles=False,
        stream_ids=stream_ids,
        force_single_stream=force_single_stream
    )

    ## over-ride timecode book making if specific values given
    if update_delay_timecodes is None and min_ctime_timecodes is None:
        min_ctime_timecodes = min_ctime
    elif min_ctime_timecodes is None:
        min_ctime_timecodes = (
            dt.datetime.now() - dt.timedelta(days=update_delay_timecodes)
        )
    if max_ctime_timecodes is None:
        max_ctime_timecodes = max_ctime

    if isinstance(min_ctime_timecodes, dt.datetime):
        min_ctime_timecodes = min_ctime_timecodes.timestamp()
    if isinstance(max_ctime, dt.datetime):
        max_ctime_timecodes = max_ctime_timecodes.timestamp()

    # hk books
    logger.info("Registering any HK Books")
    imprinter.register_hk_books(
        min_ctime=min_ctime_timecodes, 
        max_ctime=max_ctime_timecodes,
    )
    # smurf and stray books
    logger.info("Registering any timecode Books")
    imprinter.register_timecode_books(
        min_ctime=min_ctime_timecodes, 
        max_ctime=max_ctime_timecodes,
    )

    monitor = None
    if use_monitor and "monitor" in imprinter.config:
        logger.info("Will send monitor information to Influx")
        try:
            monitor = Monitor.from_configs(
                imprinter.config["monitor"]["connect_configs"]
            )
        except Exception as e:
            logger.error(f"Monitor connectioned failed {e}")
            monitor = None

    if monitor is not None:
        logger.info("Sending Updates to monitor")
        record_book_counts(monitor, imprinter)
    

def record_book_counts(monitor, imprinter):
    """Send a record of the current book count status to the InfluxDb
    site-pipeline montir
    """
    tags = [{"telescope" : imprinter.config["monitor"]["telescope"]}]
    log_tags = {}
    script_run = time.time()

    session = imprinter.get_session()
    def get_count( q ):
        return session.query(Books).filter(q).count()
    
    monitor.record(
        "unbound", 
        [ get_count(Books.status == UNBOUND) ], 
        [script_run], 
        tags, 
        imprinter.config["monitor"]["measurement"], 
        log_tags=log_tags
    )

    monitor.record(
        "bound", 
        [ get_count(Books.status == BOUND) ], 
        [script_run], 
        tags, 
        imprinter.config["monitor"]["measurement"], 
        log_tags=log_tags
    )

    monitor.record(
        "uploaded", 
        [ get_count(Books.status == UPLOADED) ], 
        [script_run], 
        tags, 
        imprinter.config["monitor"]["measurement"], 
        log_tags=log_tags
    )

    monitor.record(
        "failed", 
        [ get_count(Books.status == FAILED) ], 
        [script_run], 
        tags, 
        imprinter.config["monitor"]["measurement"], 
        log_tags=log_tags
    )

    monitor.record(
        "done", 
        [ get_count(Books.status == DONE) ], 
        [script_run], 
        tags, 
        imprinter.config["monitor"]["measurement"], 
        log_tags=log_tags
    )

    monitor.record(
        "has_level2", 
        [ get_count(not_(Books.lvl2_deleted)) ], 
        [script_run], 
        tags, 
        imprinter.config["monitor"]["measurement"], 
        log_tags=log_tags
    )

    monitor.write()

def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument('--config', type=str, help="g3tsmurf db configuration file")
    parser.add_argument('--min-ctime', type=float, help="Minimum creation time")
    parser.add_argument('--max-ctime', type=float, help="Maximum creation time")
    parser.add_argument('--stream-ids', type=str, help="Stream IDs")
    parser.add_argument(
        '--force-single-stream', help="Force single stream", action="store_true"
    )
    parser.add_argument(
        '--update-delay', type=float, 
        help="Days to subtract from now to set as minimum ctime",
        default=1
    )
    parser.add_argument(
        '--from-scratch', help="Builds or updates database from scratch",
        action="store_true"
    )
    parser.add_argument(
        '--min-ctime-timecodes', type=float, 
        help="Minimum creation time for timecode books"
    )
    parser.add_argument(
        '--max-ctime-timecodes', type=float, 
        help="Maximum creation time for timecode books"
    )
    parser.add_argument(
        '--update-delay-timecodes', type=float, 
        help= "Days to subtract from now to set as minimum ctime "
              "for timecode books",
        default=7
    )
    parser.add_argument('--use-monitor', help="Send updates to influx",
                        action="store_true")
    return parser


if __name__ == "__main__":
    parser = get_parser(parser=None)
    args = parser.parse_args()
    main(**vars(args))


"""load_obs_book

This module supports loading SO "obs/oper" books (i.e. bookbound
detector data with encoder and timestamp information) into AxisManager
data structure.

The two access points in this submodule are:

  load_obs_book

    This is a Context-compatible loader function, to be used when a
    full obsfiledb is available.

  load_book_file

    This can be used to load single G3 files (or a set of
    sample-contiguous G3 files) from a book, by filename.

Reading from g3 files can be slow on some filesystems.  Use
SOTODLIB_TOD_CACHE envvar to point to a directory where tempfiles can
be safely made, for faster acccess, such as a scratch, local, or RAM
disk.

"""

import so3g
from spt3g import core as spt3g_core
import numpy as np

from glob import glob
import contextlib
import itertools
import logging
import os
import re
import shutil
import tempfile
import yaml

import sotodlib
from sotodlib import core
from .check_book import _compact_list  # just a list with a limited repr
from . import load_smurf


logger = logging.getLogger(__name__)

_TES_BIAS_COUNT = 12  # per detset / primary file group

#: Signal DAC units are rescaled to phase before returning.
SIGNAL_RESCALE = np.pi / 2**15

DEG = np.pi / 180

TMPDIR_VAR = 'SOTODLIB_TOD_TMPDIR'


def load_obs_book(db, obs_id, dets=None, prefix=None, samples=None,
                  no_signal=None, no_headers=None, special_channels=None,
                  **kwargs):
    """Obsloader function for SO "Level 3" obs/oper Books.

    See API template, `sotodlib.core.context.obsloader_template`, for
    details of all supported arguments.

    Reading from g3 files can be slow on some filesystems.  Use
    SOTODLIB_TOD_TMPDIR envvar to point to a directory (such as a
    place on scratch server, local disk, or RAM disk) where tempfiles
    can be safely made, for faster access.  Files are created in a
    randomly named temporary directory that will (normally) be removed
    immediately after use.

    """
    if any([v is not None for v in kwargs.values()]):
        raise RuntimeError(
            f"This loader function does not understand these kwargs: f{kwargs}")

    if prefix is None:
        prefix = db.prefix
        if prefix is None:
            prefix = './'

    if no_signal is None:
        no_signal = False  # from here, assume no_signal in [True, False]
    if no_headers is None:
        # By default, suppress headers for "obs" book loads.
        no_headers = obs_id.startswith('obs_')

    # Regardless of what dets have been asked for (maybe none), get
    # the list of detsets implicated in this observation.  Make sure
    # this list is ordered by detset or the signal might not be
    # ordered properly (checked later).
    c = db.conn.execute('select distinct DS.name, DS.det from detsets DS '
                        'join files on DS.name=files.detset '
                        'where obs_id=? '
                        'order by DS.name',
                        (obs_id,))
    all_pairs = [tuple(r) for r in c.fetchall()]

    # Now filter to only the dets requested.
    if dets is None:
        pairs_req = all_pairs
    else:
        pairs_req = [p for p in all_pairs if p[1] in dets]
        dets_req = [p[1] for p in pairs_req]
        unmatched = [d for d in dets if d not in dets_req]
        if len(unmatched):
            raise RuntimeError("User requested invalid dets (e.g. %s) "
                               "for obs_id=%s" % (unmatched[0], obs_id))
        del dets_req, unmatched
    del all_pairs, dets

    # Make sure "pairs" is sorted, at _least_ at the level of grouping
    # detsets together; then make sure the detsets are processed in
    # that order.
    detsets_req = sorted(set([p[0] for p in pairs_req]))
    dets_req = []
    for _ds in detsets_req:
        dets_req.extend([p[1] for p in pairs_req if p[0] == _ds])
    del pairs_req

    # Don't pass a restriction of detsets here, as we need at least
    # one result for some downstream processing.
    file_map = db.get_files(obs_id)
    one_group = list(file_map.values())[0]  # [('file0', 0, 1000), ('file1', 1000, 2000), ...]
    
    # Figure out how many samples we're loading.
    sample_range = one_group[0][1], one_group[-1][2]
    if samples is None:
        samples = None, None
    samples = list(samples)
    if samples[0] is None:
        samples[0] = sample_range[0]
    if samples[1] is None:
        samples[1] = sample_range[1]
    elif samples[1] < 0:
        samples[1] = sample_range[1] + samples[1]
    samples[0] = min(max(0, samples[0]), sample_range[1])
    samples[1] = min(max(samples), sample_range[1])

    # Consider pre-allocating the signal buffer.
    signal_buffer = None
    if samples[1] is not None and not no_signal:
        signal_buffer = np.empty((len(dets_req), samples[1] - samples[0]),
                                 dtype='float32')

    ancil = None
    timestamps = None
    results = {}

    for detset in detsets_req:
        files = file_map[detset]
        results[detset] = _load_book_detset(
            files, prefix=prefix, load_ancil=(ancil is None),
            samples=samples, dets=dets_req, no_signal=no_signal,
            no_headers=no_headers, special_channels=special_channels,
            signal_buffer=signal_buffer)
        if ancil is None:
            ancil = results[detset]['ancil']
            timestamps = results[detset]['timestamps']

    if len(results) == 0:
        # Load the ancil files, to get ancil stuff.
        _one_fileset = next(iter(file_map.values()))
        ancil_files = _get_ancil_files(_one_fileset)
        _res = _load_book_detset(ancil_files, prefix=prefix, load_ancil=True,
                                 samples=samples, dets=[])
        ancil = _res['ancil']
        timestamps = _res['timestamps']

    if signal_buffer is not None:
        # Reduce the signal_buffer, so it only contains the loaded
        # dets. This is activated when user requests a superset of
        # dets in the files, including when the obsfiledb detset
        # includes dets that do not occur in this obs.
        loaded_dets = list(itertools.chain(*[r['dets'] for r in results.values()]))
        _, i0, i1 = core.util.get_coindices(loaded_dets, dets_req)
        assert np.array_equal(i0, np.arange(len(loaded_dets)))
        if not np.array_equal(i1, np.arange(len(signal_buffer))):
            signal_buffer = signal_buffer[i1]

    obs = _concat_filesets(results, ancil, timestamps,
                           sample0=samples[0], obs_id=obs_id,
                           signal_buffer=signal_buffer,
                           get_frame_det_info=False)
    return obs

def load_book_file(filename, dets=None, samples=None, no_signal=False,
                   no_headers=False, special_channels=False):
    """Load one or more g3 files (from an obs/oper book) and return the
    contents as an AxisManager.

    Args:

      filename (str or list of str): The g3 file(s) to load.  See
        Notes.
      dets (None, or list of str): The detectors (readout_id) to load.
      samples (tuple or None): Sample range to load.
      no_signal (bool): If True, the signal data are not read and
        .signal is set to None.
      no_headers (bool): If True, the smurf header fields are not read,
        and .primary will not be present in the output.
      special_channels (bool): If True, fixed tones (untracked)
        channels will be loaded and exposed at .tones.

    Notes:

      The filename argument can be a string or a list of strings.  If
      it's a single string, and contains a wildcard (*), then the
      pattern will be globbed and sorted and all those files will be
      loaded.

      When loading multiple files, they must be from the **same
      fileset / stream_id** (i.e. contain the same detectors).

      The "samples" argument should be a tuple (start, stop), and will
      be interpreted as sample counts within the stream of samples
      presented in the requested files, starting from 0.  The "dets"
      argument, if passed, must be a list of strings; elements in that
      list that aren't found in the data will simply be ignored (so
      the returned signal may have fewer dets than were requested).

    """
    if samples is None:
        samples = [0, None]
    else:
        samples = list(samples)
        if samples[0] is None:
            samples[0] = 0

    if isinstance(filename, str):
        if '*' in filename:
            filename = sorted(glob(filename))
        else:
            filename = [filename]

    files = [(f, None, None) for f in filename]
    this_detset = _load_book_detset(
        files, load_ancil=True,
        samples=samples, dets=dets, no_signal=no_signal,
        no_headers=no_headers, special_channels=special_channels)

    return _concat_filesets({'?': this_detset},
                            this_detset['ancil'],
                            this_detset['timestamps'],
                            sample0=samples[0])


def load_smurf_npy_data(ctx, obs_id, substr):
    """
    Loads an sodetlib npy file from Z_smurf archive of book.

    Args
    _____
    obs_id: str
        obs-id of book to load file from
    substr: str
        substring to use to find numpy file in Z_smurf
    """
    files = ctx.obsfiledb.get_files(obs_id)
    book_dir = os.path.dirname(list(files.values())[0][0][0])
    smurf_dir = os.path.join(book_dir, 'Z_smurf')
    for f in os.listdir(smurf_dir):
        if substr in f:
            fpath = os.path.join(smurf_dir, f)
            break
    else:
        raise FileNotFoundError("Could not find npy file")
    res = np.load(fpath, allow_pickle=True).item()
    return res


def _load_book_detset(files, prefix='', load_ancil=True,
                      dets=None, samples=None, no_signal=False,
                      no_headers=False, special_channels=False,
                      signal_buffer=None):
    """Read data from a single detset.

    If a list of dets is specified, it may include dets that aren't
    found in this set of files.  It is thus used only to screen
    detectors and to set the order they appear in the output.  The
    actual detectors found and loaded are returned as 'dets' in the
    output.

    """
    stream_id = None
    ancil_acc = None
    times_acc = None

    flag_accs = {}
    this_stream_dets = None

    if load_ancil:
        times_acc = Accumulator1d(samples=samples)
        ancil_acc = AccumulatorTimesampleMap(samples=samples)
    if no_headers:
        primary_acc = None
        bias_names = None
        bias_acc = None
    else:
        primary_acc = AccumulatorNamed(samples=samples)
        bias_names = []
        bias_acc = Accumulator2d(samples=samples)

    if no_signal:
        signal_acc = None
    elif signal_buffer is not None:
        signal_acc = Accumulator2d(
            samples=samples,
            insert_at=signal_buffer,
            keys_to_keep=dets,
            calibrate=SIGNAL_RESCALE)
    else:
        signal_acc = Accumulator2d(
            samples=samples,
            keys_to_keep=dets,
            calibrate=SIGNAL_RESCALE)

    if special_channels:
        tones_acc = Accumulator2d(
            samples=samples,
            calibrate=SIGNAL_RESCALE)
    else:
        tones_acc = None

    if (not no_signal) or (not no_headers) or (special_channels):
        # If we're not otherwise scanning every frame ... don't due
        # any smurfy flags.
        flag_accs['smurfgaps'] = Accumulator1d(samples=samples)

    # Sniff out a smurf status frame.
    smurf_proc = load_smurf.SmurfStatus._get_frame_processor()

    for frame, frame_offset in _frames_iterator(files, prefix, samples,
                                                smurf_proc=smurf_proc):
        # This is to escape once requested number of samples (and a
        # smurf dump frame) are read.  Loop can exit early if no data
        # are actually required from frame.
        more_data = (not smurf_proc.get('dump_frame', False))

        # Anything in ancil should be identical across
        # filesets, so only process it once.
        if load_ancil:
            more_data |= times_acc.append(frame['ancil'].times, frame_offset)
            more_data |= ancil_acc.append(frame['ancil'], frame_offset)

        if 'stream_id' in frame:
            if stream_id is None:
                stream_id = frame['stream_id']
            assert (stream_id == frame['stream_id'])  # check your data files

        if 'primary' in frame and primary_acc is not None:
            more_data |= primary_acc.append(frame['primary'], frame_offset)
            bias_names = _check_bias_names(frame)[:_TES_BIAS_COUNT]
            more_data |= bias_acc.append(frame['tes_biases'], frame_offset)

        if 'signal' in frame:
            # Even if no_signal, we need the det list.
            if this_stream_dets is None:
                this_stream_dets = _compact_list(frame['signal'].names)

            # Extract the main signal
            if not no_signal:
                more_data |= signal_acc.append(frame['signal'], frame_offset)

        if 'untracked' in frame and special_channels:
            more_data |= tones_acc.append(frame['untracked'], frame_offset)

        if 'flag_smurfgaps' in frame and 'smurfgaps' in flag_accs:
            more_data |= flag_accs['smurfgaps'].append(frame['flag_smurfgaps'])

        if not more_data:
            break

    if times_acc is not None:
        times_acc = times_acc.finalize() / spt3g_core.G3Units.sec

    req_dets_in_stream = None
    det_idx_in_stream = None
    if this_stream_dets:
        # Of the requested detectors, what ones were actually found in
        # this stream?  Keep ordering same as the request, which
        # should be the same as the populated data.
        if dets is None:
            req_dets_in_stream = this_stream_dets
        else:
            req_dets_in_stream = _compact_list(
                [d for d in dets if d in this_stream_dets])

        # For each loaded detector, what was its index within the stream's
        # dets?
        det_idx_in_stream = _compact_list([this_stream_dets.index(d)
                                           for d in req_dets_in_stream])

    stat = smurf_proc.get_status()
    ch_info = None
    iir_params = None
    if stat.num_chans is None:
        # Try to grab it from file 000?
        try_file = re.sub(r'_\d\d\d\.g3', '_000.g3', files[0][0])
        if try_file != files[0][0] and os.path.exists(try_file):
            logger.warning(f'Trying to get SmurfStatus from {try_file} ...')
            stat = load_smurf.SmurfStatus.from_file(try_file)
            if stat.num_chans is None:
                logger.warning('... it did not work.')
                # No wiring frames probably means it's an A_* (ancil) file.

    if stat.num_chans is not None:
        # This is an AxisManager, with dets axis for just this stream
        # ... extract and stack data later.
        ch_info = load_smurf.get_channel_info(stat, mask=det_idx_in_stream)
        # And this stuff is per stream, so keep it separate.
        iir_params = {'enabled': stat.filter_enabled,
                      'b': stat.filter_b,
                      'a': stat.filter_a,
                      'fscale': 1. / stat.flux_ramp_rate_hz}

    return {
        'stream_id': stream_id,
        'signal': signal_acc,
        'dets': req_dets_in_stream,
        'primary': primary_acc,
        'biases': bias_acc,
        'bias_names': bias_names,
        'smurf_ch_info': ch_info,
        'iir_params': iir_params,
        'tones': tones_acc,
        'flags': flag_accs,
        'ancil': ancil_acc,
        'timestamps': times_acc,
    }


def _concat_filesets(results, ancil=None, timestamps=None,
                     sample0=0, obs_id=None, dets=None,
                     signal_buffer=None,
                     get_frame_det_info=True):
    """Assemble multiple detset results (as returned by _load_book_detset)
    into a full AxisManager.

    """
    if ancil is None:
        ancil = next(iter(results))['ancil']
    if timestamps is None:
        timestamps = next(iter(results))['timestamps']

    if dets is None:
        dets = list(itertools.chain(*[r['dets'] for r in results.values()]))

    aman = core.AxisManager(
        core.LabelAxis('dets', dets),
        core.OffsetAxis('samps',
                        count=len(timestamps),
                        offset=sample0,
                        origin_tag=obs_id))

    aman.wrap('timestamps', timestamps, axis_map=[(0, 'samps')])

    if ancil is not None:
        # Handle ancillary fields, a.k.a. boresight pointing /
        # rotation / corotator.
        aman.wrap('ancil', core.AxisManager(aman.samps))

        # Put all fields into 'ancil'.
        _a = aman['ancil']
        for k, v in ancil.finalize().items():
            _a.wrap(k, v, [(0, 'samps')])

        # Transform some fields into 'boresight'.
        if 'az_enc' in _a and 'el_enc' in _a:
            aman.wrap('boresight', core.AxisManager(aman.samps))
            _b = aman['boresight']
            _b.wrap('az', _a['az_enc'] * DEG, [(0, 'samps')])
            _b.wrap('el', _a['el_enc'] * DEG, [(0, 'samps')])

            roll = None
            if 'boresight_enc' in _a:
                roll = -1*_a['boresight_enc']
            elif 'corotator_enc' in _a:
                roll = _a['el_enc'] - 60 - _a['corotator_enc']
            if roll is None:
                _b.wrap('roll', None)
            else:
                _b.wrap('roll', roll * DEG, [(0, 'samps')])

    if len(results) == 0:
        return aman

    one_result = next(iter(results.values()))
    no_signal = one_result['signal'] is None
    if no_signal:
        aman.wrap('signal', None)
    else:
        if signal_buffer is not None:
            aman.wrap('signal', signal_buffer,
                      [(0, 'dets'), (1, 'samps')])
        else:
            aman.wrap_new('signal', shape=('dets', 'samps'), dtype='float32')
            dets_ofs = 0
            for v in results.values():
                d = v['signal'].finalize()
                aman['signal'][dets_ofs:dets_ofs + len(d)] = d
                dets_ofs += len(d)

    if one_result['tones'] is not None:
        # Expose the "untracked" channels, i.e. fixed tones.
        n_tones = 0
        tone_info = []
        for v in results.values():
            d = v['tones'].finalize()
            n_tones += d.shape[0]
            for k in v['tones'].keys:
                # Should look like this: sch_NONE_2_326
                b, c = map(int, k.split('_')[2:])
                tone_info.append((v['stream_id'], v['stream_id'] + f'_{b}_{c}', b, c))
        ts, tk, tb, tc = map(np.array, zip(*tone_info))
        tman = core.AxisManager(core.LabelAxis('tdets', tk),
                                aman.samps)
        aman.wrap('tones', tman)
        aman.tones.wrap('stream_id', ts, axis_map=[(0, 'tdets')])
        aman.tones.wrap('band', tb, axis_map=[(0, 'tdets')])
        aman.tones.wrap('channel', tc, axis_map=[(0, 'tdets')])
        aman.tones.wrap_new('signal', shape=('tdets', 'samps'), dtype='float32')
        dets_ofs = 0
        for v in results.values():
            d = v['tones'].data
            aman.tones['signal'][dets_ofs:dets_ofs + len(d)] = d
            dets_ofs += len(d)

    # In sims, or if no_headers, the primary block may be unpopulated.
    if any([(v['primary'] is not None and v['primary'].data is not None)
            for v in results.values()]):
        # Biases
        all_bias_names = []
        for v in results.values():
            all_bias_names.extend(v['bias_names'][:_TES_BIAS_COUNT])
        aman.merge(core.AxisManager(core.LabelAxis('bias_lines', all_bias_names)))
        aman.wrap_new('biases', shape=('bias_lines', 'samps'), dtype='int32')
        for i, v in enumerate(results.values()):
            aman['biases'][i * _TES_BIAS_COUNT:(i + 1) * _TES_BIAS_COUNT, :] = \
                v['biases'].finalize()[:_TES_BIAS_COUNT, :]

        # Other header data, per-stream
        aman.wrap('primary', core.AxisManager(aman.samps))
        for r in results.values():
            # Primary.
            _prim = core.AxisManager(aman.samps)
            for k, v in r['primary'].finalize().items():
                _prim.wrap(k, v, [(0, 'samps')])
            aman['primary'].wrap(r['stream_id'], _prim)


    if any([v['iir_params'] is not None
            for v in results.values()]):
        # Filter parameters, per-stream
        aman.wrap('iir_params', core.AxisManager())
        aman['iir_params'].wrap('per_stream', True)
        for r in results.values():
            _iir = None
            if r.get('iir_params') is not None:
                _iir = core.AxisManager()
                for k, v in r['iir_params'].items():
                    _iir.wrap(k, v)
            aman['iir_params'].wrap(r['stream_id'], _iir)

    # flags place
    aman.wrap("flags", core.FlagManager.for_tod(aman, "dets", "samps"))

    # Assemble per-wafer flags by suffixing the stream_id.
    for flag_key in one_result['flags'].keys():
        for v in results.values():
            fasr = so3g.RangesInt32.from_mask(v['flags'][flag_key].finalize())
            aman.flags.wrap(f'{flag_key}_{v["stream_id"]}', fasr, axis_map=[(0, 'samps')])

    if not get_frame_det_info:
        return aman

    # The detset, stream_id, and smurf.* channel info will normally be
    # populated by a downstream data product, so that they are
    # available without having to read the main G3 data (and thus with
    # get_meta).  But the block below should be maintained for use
    # with load_book_file, where the user is unlikely to also have
    # good metadata ready to go.
    #
    # Even if the smurf info isn't merged in here, it still gets
    # parsed.  The main need seems to be to populate the iir_params.

    # det_info
    det_info = core.metadata.ResultSet(
        ['detset', '_readout_id', 'stream_id'])
    ch_info = None  # or False, or dict.
    for detset, r in results.items():
        det_info.rows.extend(
            [(detset, _d, r['stream_id']) for _d in r['dets']])

        if r['smurf_ch_info'] is None:
            ch_info = False
            break

        _ch_info_keys = list(r['smurf_ch_info']._fields.keys())

        if ch_info is None:
            ch_info = {k: [] for k in _ch_info_keys}
        elif set(ch_info.keys()) != set(_ch_info_keys):
            ch_info = False
            break

    if ch_info is False:
        logger.warning('Missing or inconsistent smurf status fields; '
                       'dropping det_info.smurf.')

    if ch_info:
        for detset, r in results.items():
            if r['smurf_ch_info'] is None:
                break
            for k, v in r['smurf_ch_info']._fields.items():
                if k not in ch_info:
                    ch_info[k] = []
                ch_info[k].extend(v)

    aman.wrap('det_info', det_info.to_axismanager(axis_key='_readout_id'))

    if ch_info:
        smurf = core.AxisManager(aman.dets)
        for k, v in ch_info.items():
            smurf.wrap(k, np.array(v), [(0, 'dets')])
        aman['det_info'].wrap('smurf', smurf)

    return aman


def _check_bias_names(frame):
    """Verify the frame has TES biases with expected names; return the
    modified names that include stream_id.

    """
    for i, name in enumerate(frame['tes_biases'].names):
        if name != 'bias%02i' % i:
            raise RuntimeError(f'Bias at index {i} has unexpected name "{name}"!')
    stream_id = frame['stream_id']
    return [f'{stream_id}_b{_i:02d}' for _i in range(i+1)]


def _get_ancil_files(non_ancil_files):
    def _rewrite_entry(row):
        f, etc = row[0], row[1:]
        p, b = os.path.split(f)
        tokens = b.split('_')
        a = os.path.join(p, 'A_ancil_' + tokens[-1])
        return tuple([a] + list(etc))
    return [_rewrite_entry(x) for x in non_ancil_files]


class Accumulator:
    def __init__(self, shape=None, samples=None, preconsumed=None):
        if samples is None:
            samples = None, None
        samples = list(samples)
        if samples[0] == None:
            samples[0] = 0
        self.samples = samples

        self.shape = None
        self.data = None
        self.consumed = preconsumed

    def append(self, data, preconsumed=None):
        if self.consumed is None:
            if preconsumed is None:
                preconsumed = 0
            self.consumed = preconsumed
        if preconsumed is not None:
            assert(self.consumed == preconsumed)

        data_count = self._sample_count(data)

        # global sample indices of our destination buffer are
        # self.samples; global sample indices of this block are:
        block_samples = [self.consumed, self.consumed + data_count]

        # Does this block precede our area of interest?
        if block_samples[1] <= self.samples[0]:
            self.consumed += data_count
            return True

        # Overlap?
        over_samples = [max(block_samples[0], self.samples[0]),
                        block_samples[1]]
        if self.samples[1] is not None:
            over_samples[1] = min(over_samples[1], self.samples[1])

        # Size of overlap
        if over_samples[1] - over_samples[0] <= 0:
            return False

        # Extraction slice
        src_slice = slice(over_samples[0] - self.consumed,
                          over_samples[1] - self.consumed)

        # Insertion slice
        dest_slice = slice(over_samples[0] - self.samples[0],
                           over_samples[1] - self.samples[0])

        # Specialization ...
        self._extract(data, src_slice, dest_slice)

        if over_samples[1] != block_samples[1]:
            return False

        self.consumed += data_count
        return True


class Accumulator1d(Accumulator):
    def _sample_count(self, _data):
        return len(_data)

    def _extract(self, data, src_slice, dest_slice):
        _data = np.asarray(data[src_slice])

        # On first frame, check if we know the final data shape.
        if self.data is None:
            if self.samples[1] is not None:
                self.shape = (self.samples[1] - self.samples[0], )

        if self.shape is not None:
            # Determinate.
            if self.data is None:
                self.data = np.empty(self.shape, dtype=_data.dtype)
            self.data[dest_slice] = _data
        else:
            # Indeterminate
            if self.data is None:
                self.data = []
            self.data.append(_data)

    def finalize(self):
        if self.shape is None:
            return np.hstack(self.data)
        else:
            return self.data


class AccumulatorNamed(Accumulator):
    """Accumulator for unpacking 2-d data (G3SuperTimestream) into
    individual (named) 1-d vectors.

    """
    def _sample_count(self, _data):
        return len(_data.times)

    def _extract(self, data, src_slice, dest_slice):
        # On first frame, check if we know the final data shape.
        if self.data is None:
            if self.samples[1] is not None:
                self.shape = (self.samples[1] - self.samples[0], )
            if hasattr(data, 'names'):
                # G3SuperTimestream ...
                self.keys = [k for k in data.names]
            else:
                # G3TimesampleMap ...
                self.keys = [k for k in data.keys()]

        if self.shape is not None:
            # Determinate.
            if self.data is None:
                self.data = {k: np.empty(self.shape[-1], dtype=data.data.dtype)
                             for k in self.keys}
            for i, k in enumerate(self.keys):
                self.data[k][dest_slice] = data.data[i][src_slice]
        else:
            # Indeterminate
            if self.data is None:
                self.data = {k: [] for k in self.keys}
            for i, k in enumerate(self.keys):
                self.data[k].append(data.data[i][src_slice])

    def finalize(self):
        if self.shape is None:
            self.data = {k: np.hstack(self.data[k]) for k in self.keys}
        return self.data


class AccumulatorTimesampleMap(AccumulatorNamed):
    """Accumulator for unpacking 2-d data (G3TimestampleMap) into
    individual (named) 1-d vectors.

    """
    def _extract(self, data, src_slice, dest_slice):
        # On first frame, check if we know the final data shape.
        if self.data is None:
            if self.samples[1] is not None:
                self.shape = (self.samples[1] - self.samples[0], )
            self.keys = [k for k in data.keys()]

        if self.shape is not None:
            # Determinate.
            if self.data is None:
                self.data = {k: np.empty(self.shape[-1], dtype=np.asarray(data[k]).dtype)
                             for k in self.keys}
            for i, k in enumerate(self.keys):
                self.data[k][dest_slice] = data[k][src_slice]
        else:
            # Indeterminate
            if self.data is None:
                self.data = {k: [] for k in self.keys}
            for i, k in enumerate(self.keys):
                self.data[k].append(data[k][src_slice])


class Accumulator2d(Accumulator):
    """Accumulator for unpacking 2-d data (G3SuperTimestream) into
    2-d array (preserving first axis labels).

    """
    def __init__(self, *args, insert_at=None, keys_to_keep=None,
                 calibrate=None, **kwargs):
        super().__init__(*args, **kwargs)
        # An optional destination buffer for the data.
        self.insert_at = insert_at
        self.keys_to_keep = keys_to_keep
        self.insert_at_idx = None
        self.extract_at_idx = None
        if self.insert_at is not None and self.samples[1] is None:
            self.samples[1] = self.insert_at.shape[-1] + self.samples[0]
        self.calibrate = calibrate

    def _sample_count(self, _data):
        return len(_data.times)

    def _extract(self, data, src_slice, dest_slice):

        if self.calibrate is not None:
            # This is a low cost operation if you do it before
            # decompression.  (Also do it before you use data.dtype,
            # in "first frame stuff".)
            data.calibrate(np.array([self.calibrate] * len(data.names)))

        # First frame stuff ...
        if self.data is None:
            if self.insert_at is not None:
                # We have a destination buffer
                self.keys, self.extract_at_idx, self.insert_at_idx = \
                    core.util.get_coindices(
                        data.names, self.keys_to_keep)
                self.data = self.insert_at  # place-holder
            else:
                if self.keys_to_keep is not None:
                    self.keys, _, self.extract_at_idx = \
                        core.util.get_coindices(self.keys_to_keep, data.names)
                else:
                    self.keys = list(data.names)
                # Set shape, if we know it.
                if self.samples[1] is not None:
                    self.shape = (len(self.keys), self.samples[1] - self.samples[0])
                    self.data = np.empty(self.shape, data.dtype)
                else:
                    self.data = []

        # G3SuperTimestream.extract() is available from so3g v0.1.13
        # (April 2024).  The previous handling (below this block) can
        # be removed in a few months.
        if hasattr(data, 'extract'):
            if self.insert_at is not None:
                data.extract(self.insert_at[:, dest_slice],
                             self.insert_at_idx,
                             self.extract_at_idx,
                             src_slice.start, src_slice.stop)
            elif self.shape is not None:
                if len(data.names) > 0:
                    data.extract(self.data[:, dest_slice], None, self.extract_at_idx,
                                 src_slice.start, src_slice.stop)
            else:
                _sh = [len(data.names), len(data.times)]
                if self.extract_at_idx is not None:
                    _sh[0] = len(self.extract_at_idx)
                _dest = np.empty(_sh, dtype=data.dtype)
                data.extract(_dest, None, self.extract_at_idx,
                             src_slice.start, src_slice.stop)
                self.data.append(_dest)
            return

        # Store data from this frame.
        if self.insert_at is not None:
            # Indexed by name
            for i0, i1 in zip(self.insert_at_idx, self.extract_at_idx):
                self.insert_at[i0, dest_slice] = data.data[i1, src_slice]

        elif self.shape is not None:
            # Full array to hold data.
            if self.extract_at_idx is not None:
                for i, j in enumerate(self.extract_at_idx):
                    self.data[i, dest_slice] = data.data[j, src_slice]
            else:
                self.data[:, dest_slice] = data.data[:, src_slice]

        else:
            # List of arrays, to be hstacked later.
            if self.extract_at_idx is not None:
                self.data.append(data.data[self.extract_at_idx, src_slice])
            else:
                self.data.append(data.data[:, src_slice])

    def finalize(self):
        if self.insert_at is not None:
            pass
        elif self.shape is None:
            self.data = np.hstack(self.data)
        return self.data


def _frames_iterator(files, prefix, samples, smurf_proc=None, use_temp_dir=None):
    """Iterates over frames in files.  yields only frames that might be of
    interest for timestream unpacking.

    Yields each (frame, offset).  The offset is the global offset
    associated with the start of the frame.

    If use_temp_dir is a string, data files are copied to that dir
    before reading from them.  If use_temp_dir is None, and
    SOTODLIB_TOD_CACHE envvar is defined, then a TemporaryDirectory
    (which can clean up after itself) will be created inside
    SOTODLIB_TOD_CACHE and that will be used as the temp_dir.

    """
    tmpdir_root = os.getenv(TMPDIR_VAR)
    if tmpdir_root:
        cmgr = tempfile.TemporaryDirectory(dir=tmpdir_root)
    else:
        # This yields None
        cmgr = contextlib.nullcontext()

    with cmgr as tmpdir:
        offset = 0
        for f, i0, i1 in files:
            if i0 is None:
                i0 = offset

            if smurf_proc is not None:
                if (i1 is not None) and (i1 <= samples[0]):
                    continue
                if samples[1] is not None and i0 >= samples[1]:
                    break

            filename = os.path.join(prefix, f)
            offset = i0

            if tmpdir:
                filename, orig_filename = os.path.join(tmpdir, 'framefile'), filename
                logger.debug('Copying data file %s to %s.' % (orig_filename, filename))
                shutil.copyfile(orig_filename, filename)

            for frame in spt3g_core.G3File(filename):
                if smurf_proc is not None and smurf_proc.process(frame):
                    # We found a dump frame, so stop looking.
                    smurf_proc = None
                if frame.type is not spt3g_core.G3FrameType.Scan:
                    continue
                yield frame, offset
                offset += len(frame['ancil'].times)
                # Alternately, use frame['sample_range']


def get_cal_obsids(ctx, obs_id, cal_type):
    """
    Returns set of obs-ids corresponding to the most recent calibration
    operations for a given obsid.

    Args
    ------
    ctx: core.Context
        Context object
    obs_id: str
        obs_id for which you want to get relevant calibration info
    cal_type: str
        Calibration subtype to use in the obsdb query. For example: 'iv' or
        'bias_steps'.

    Returns
    ----------
        obs_ids: dict
            Dict of obs_ids for each detset in specified operation
    """
    obs = ctx.obsdb.query(f"obs_id == '{obs_id}'")[0]
    detsets = ctx.obsfiledb.get_detsets(obs_id)
    min_ct = obs['start_time'] - 3600*24*7
    cal_all = ctx.obsdb.query(
        f"""
        start_time <= {obs['start_time']} and subtype=='{cal_type}'
        and start_time > {min_ct}
        """, sort=['start_time']
    )[::-1]

    obs_ids = {
        ds: None for ds in detsets
    }
    ids_to_find = len(obs_ids)
    ids_found = 0

    for o in cal_all:
        dsets = ctx.obsfiledb.get_files(o['obs_id']).keys()
        for ds in dsets:
            if ds in obs_ids:
                if obs_ids[ds] is None:
                    obs_ids[ds] = o['obs_id']
                    ids_found += 1
        if ids_to_find == ids_found:
            break

    return obs_ids


core.OBSLOADER_REGISTRY['obs-book'] = load_obs_book



from .utils import *
from .ml_mapmaker import *
from .demod_mapmaker import *
from .noise_model import *
from .pointing_matrix import *
from .obs_grouping import *
from .bench import *
from .log import *



# Copyright (c) 2018-2020 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Data Conditioning

This module contains code for conditioning G3Timestream data in G3Frames

"""
import numpy as np
import scipy.signal as signal

from spt3g import core
from .core.g3_core import DataG3Module

class MeanSubtract(DataG3Module):
    def process(self, data, det_name):
        return data-np.nanmean(data)

class MedianSubtract(DataG3Module):
    def process(self, data, det_name):
        return data-np.nanmedian(data)

class Detrend(DataG3Module):
    """
    Module for Detrending data. Information is added to the frame so that the
    resulting data can be easily re-trended (if, for example the detrend is done
    just for filtering).
    """

    def __init__(self, input='signal', output=None,
                        info='detrend_values', type='linear'):
        """
        Args:
            info (str): key for where the values will be saved
            type (str): can be 'constant' or 'linear'
        """
        self.type=type
        if self.type=='linear':
            self.deg=1
        elif self.type=='constant':
            self.deg=0
        else:
            raise ValueError("type must be 'linear' or 'constant'")
        self.info = info
        super().__init__(input, output)

    def __call__(self, f):
        if f.type == core.G3FrameType.Scan:
            self.detrend_vals = core.G3MapVectorDouble()

        super().__call__(f)

        if f.type == core.G3FrameType.Scan:
            f[self.info] = self.detrend_vals

    def process(self, data, det_name):
        x=np.arange(data.n_samples)
        self.detrend_vals[det_name] = np.polyfit(x, data, deg=self.deg)

        return data - np.polyval(self.detrend_vals[det_name], x)

class Retrend(DataG3Module):
    """
    Module for Retrending data that was Detrended with Detrend
    """
    def __init__(self, input='signal', output=None,
                        detrend_info='detrend_values'):
        """
        Args:
            info (str): key for where the values from detrending are saved
        """

        self.info = detrend_info
        super().__init__(input, output)

    def __call__(self, f):
        if f.type == core.G3FrameType.Scan:
            if self.info not in f.keys():
                raise ValueError('No Detrending information in {}'.format(self.info))
            else:
                self.retrend = f[self.info]

        super().__call__(f)

        if f.type == core.G3FrameType.Scan:
            f.pop(self.info)

    def process(self, data, det_name):
        x=np.arange(data.n_samples)
        return data + np.polyval(self.retrend[det_name], x)

class Decimate(DataG3Module):
    """
    Module for decimating data. Uses scipy.signal.decimate()
    """
    def __init__(self, input='signal', output=None, q=5, **kwargs):
        """
        Arguments:
            q (int): The downsampling factor
            kwargs: can include any of the optional parameters for
                scipy.signal.decimate
        """
        self.decimate_params = {'q': q, 'zero_phase': True}
        self.decimate_params.update(kwargs)

        super().__init__(input, output)

    def process(self, data, det_name):
        return signal.decimate(data, **self.decimate_params)

class Resample(DataG3Module):
    """
    Module for resampling data. Uses scipy.signal.resample()
    """
    def __init__(self, input='signal', output=None, num=3000, **kwargs):
        """
        Arguments:
            num (int): The number of samples in the resampled signal.
            kwargs: can include any of the optional parameters for
                scipy.signal.resample
        """
        self.resample_params = {'num':num}
        self.resample_params.update(kwargs)

        super().__init__(input, output)

    def process(self, data, det_name):
        return signal.resample(data, **self.resample_params)



# Copyright (c) 2023-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simultaneous filtering and binned mapmaking.
"""

import numpy as np
from astropy import units as u
import toast
import toast.ops
from toast.observation import default_values as defaults

from .. import ops as so_ops
from .job import workflow_timer


def setup_mapmaker_filterbin(operators):
    """Add commandline args, operators, and templates for TOAST FilterBin mapmaker.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.FilterBin(name="filterbin", enabled=False))


@workflow_timer
def mapmaker_filterbin(job, otherargs, runargs, data):
    """Run the TOAST FilterBin mapmaker.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    log = toast.utils.Logger.get()

    # Configured operators for this job
    job_ops = job.operators

    if job_ops.filterbin.enabled:
        job_ops.filterbin.binning = job_ops.binner_final
        job_ops.filterbin.binning.full_pointing = otherargs.full_pointing
        job_ops.filterbin.det_data = job_ops.sim_noise.det_data
        job_ops.filterbin.output_dir = otherargs.out_dir
        if otherargs.obsmaps:
            # Map each observation separately
            timer_obs = toast.timing.Timer()
            timer_obs.start()
            group = data.comm.group
            orig_name_filterbin = job_ops.filterbin.name
            orig_comm = data.comm
            new_comm = toast.mpi.Comm(world=data.comm.comm_group)
            for iobs, obs in enumerate(data.obs):
                log.info_rank(
                    f"{group} : mapping observation {iobs + 1} / {len(data.obs)}.",
                    comm=new_comm.comm_world,
                )
                # Data object that only covers one observation
                obs_data = data.select(obs_uid=obs.uid)
                # Replace comm_world with the group communicator
                obs_data._comm = new_comm
                job_ops.filterbin.name = f"{orig_name_filterbin}_{obs.name}"
                job_ops.filterbin.reset_pix_dist = True
                job_ops.filterbin.apply(obs_data)
                log.info_rank(
                    f"{group} : Filter+binned {obs.name} in",
                    comm=new_comm.comm_world,
                    timer=timer_obs,
                )
            log.info_rank(
                f"{group} : Done mapping {len(data.obs)} observations.",
                comm=new_comm.comm_world,
            )
            data._comm = orig_comm
        else:
            job_ops.filterbin.apply(data)



import os
import yaml
import time
import logging
import numpy as np
import argparse
import traceback
from typing import Optional
import copy

from sotodlib import core
import sotodlib.preprocess.preprocess_util as pp_util
from sotodlib.preprocess import _Preprocess, Pipeline, processes

logger = sp_util.init_logger("preprocess")

def plot_preprocess_tod(obs_id, configs, context, group_list=None, verbosity=2):
    """ Loads the saved information from the preprocessing pipeline and runs the
    plotting section of the pipeline. 

    Assumes preprocess_tod has already been run on the requested observation. 
    
    Arguments
    ----------
    obs_id: multiple
        passed to `context.get_obs` to load AxisManager, see Notes for 
        `context.get_obs`
    configs: string or dictionary
        config file or loaded config directory
    """
    logger = sp_util.init_logger("preprocess", verbosity=verbosity)
    
    group_by, groups = pp_util.get_groups(obs_id, configs, context)
    all_groups = groups.copy()
    for g in all_groups:
        if group_list is not None:
            if g not in group_list:
                groups.remove(g)
                continue
        if 'wafer.bandpass' in group_by:
            if 'NC' in g:
                groups.remove(g)
                continue
        try:
            meta = context.get_meta(obs_id, dets = {gb:gg for gb, gg in zip(group_by, g)})
        except Exception as e:
            errmsg = f'{type(e)}: {e}'
            tb = ''.join(traceback.format_tb(e.__traceback__))
            logger.info(f"ERROR: {obs_id} {g}\n{errmsg}\n{tb}")
            groups.remove(g)
            continue

        if meta.dets.count == 0:
            groups.remove(g)
        
    if len(groups) == 0:
        logger.warning(f"group_list:{group_list} contains no overlap with "
                       f"groups in observation: {obs_id}:{all_groups}. "
                       f"No analysis to run.")
        return
    
    pipe = Pipeline(configs["process_pipe"], plot_dir=configs["plot_dir"], logger=logger)
    
    for group in groups:
        logger.info(f"Beginning run for {obs_id}:{group}")
        try:
            meta = context.get_meta(obs_id, dets={gb:g for gb, g in zip(group_by, group)})
            aman = context.get_obs(meta)
            pipe.run(aman, aman.preprocess, update_plot=True)
        except Exception as e:
            errmsg = f'{type(e)}: {e}'
            tb = ''.join(traceback.format_tb(e.__traceback__))
            logger.info(f"ERROR: {obs_id} {group}\n{errmsg}\n{tb}")
            continue

def plot_preprocess_tod_from_db(obs_id, group_by, group, configs, context, verbosity=2):
    """ Loads the saved information from the preprocessing pipeline and runs the
    plotting section of the pipeline. 

    Assumes preprocess_tod has already been run on the requested observation. 
    
    Arguments
    ----------
    obs_id: multiple
        passed to `context.get_obs` to load AxisManager, see Notes for 
        `context.get_obs`
    configs: string or dictionary
        config file or loaded config directory
    """
    logger = sp_util.init_logger("preprocess", verbosity=verbosity)
    
    pipe = Pipeline(configs["process_pipe"], plot_dir=configs["plot_dir"], logger=logger)
    
    logger.info(f"Beginning run for {obs_id}:{group}")
    try:
        meta = context.get_meta(obs_id, dets={group_by: group})
        aman = context.get_obs(meta)
        pipe.run(aman, aman.preprocess, update_plot=True)
    except Exception as e:
        errmsg = f'{type(e)}: {e}'
        tb = ''.join(traceback.format_tb(e.__traceback__))
        logger.info(f"ERROR: {obs_id} {group}\n{errmsg}\n{tb}")


def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument('configs', help="Preprocessing Configuration File")
    parser.add_argument(
        '--query', 
        help="Query to pass to the observation list. Use \\'string\\' to "
             "pass in strings within the query.",  
        type=str
    )
    parser.add_argument(
        '--use-db', 
        help="If true, get obs list from existing database provided in config",  
        action='store_true',
    )
    parser.add_argument(
        '--obs-id',
        help="obs-id of particular observation if we want to run on just one"
    )
    parser.add_argument(
        '--overwrite',
        help="If true, overwrites existing entries in the database",
        action='store_true',
    )
    parser.add_argument(
        '--min-ctime',
        help="Minimum timestamp for the beginning of an observation list",
    )
    parser.add_argument(
        '--max-ctime',
        help="Maximum timestamp for the beginning of an observation list",
    )
    parser.add_argument(
        '--update-delay',
        help="Number of days (unit is days) in the past to start observation list.",
        type=int
    )
    parser.add_argument(
        '--tags',
        help="Observation tags. Ex: --tags 'jupiter' 'setting'",
        nargs='*',
        type=str
    )
    parser.add_argument(
        '--planet-obs',
        help="If true, takes all planet tags as logical OR and adjusts related configs",
        action='store_true',
    )
    parser.add_argument(
        '--verbosity',
        help="increase output verbosity. 0:Error, 1:Warning, 2:Info(default), 3:Debug",
        default=2,
        type=int
    )
    return parser

def main(
        configs: str,
        query: Optional[str] = None, 
        use_db: bool = False, 
        obs_id: Optional[str] = None, 
        overwrite: bool = False,
        min_ctime: Optional[int] = None,
        max_ctime: Optional[int] = None,
        update_delay: Optional[int] = None,
        tags: Optional[str] = None,
        planet_obs: bool = False,
        verbosity: Optional[int] = None,
 ):
    """The intended use case of this script is: if preprocessing has been previously
    run on a set of observations but plotting was not enabled during that run. Now,
    we re-run preprocessing on those observations to produce the plots, without having
    to fully re-run all the calc steps. Provide the preprocess config file that was
    initially used, this time enabling plotting in the pipeline where necessary. The
    config should point to the existing database with the preprocess data. Set use_db=True
    to get the obs list and groups from the existing database (Note: this will also set
    overwrite=True). Otherwise, you can provide a query similiar to the query used for
    preprocess_tod to define the obs you want to run plotting on.

    """
    configs, context = pp_util.get_preprocess_context(configs)
    logger = sp_util.init_logger("preprocess", verbosity=verbosity)

    if not os.path.exists(configs['archive']['index']):
        # don't run if database doesn't exist
        raise Exception(f"No database found at {configs['archive']['index']}.")
    else:
        db = core.metadata.ManifestDb(configs['archive']['index'])

    run_list = []

    if use_db:
        overwrite = True
        obs_list = db.inspect()
        for obs in obs_list:
            obsid = obs['obs:obs_id']
            for k, v in obs.items():
                if k.startswith('dets:'):
                    group_by = k.split(':',1)[1]
                    group = v
            run_list.append( ({'obs_id': obsid, 'group_by': group_by}, group) )
    else:
        obs_list = sp_util.get_obslist(context, query=query, obs_id=obs_id, min_ctime=min_ctime, 
                                       max_ctime=max_ctime, update_delay=update_delay, tags=tags, 
                                       planet_obs=planet_obs)
        if len(obs_list)==0:
            logger.warning(f"No observations returned from query: {query}")

        for obs in obs_list:
            x = db.inspect({'obs:obs_id': obs["obs_id"]})
            if x is None or len(x) == 0:
                logger.warning(f"Obs_id {obs['obs_id']} not found in database.")
            else:
                run_list.append( (obs, None) )

    logger.info(f"Beginning to run preprocessing on {len(run_list)} observations")
    for obs, groups in run_list:
        obsid = obs["obs_id"]
        ctime = obsid[4:9]
        plot_path = os.path.join(configs['plot_dir'], f'{ctime}/{obsid}')
        logger.info(f"Processing obs_id: {obsid}")
        if not overwrite and os.path.exists(plot_path):
            logger.warning(f"Plots found at {plot_path}. Set overwrite=True to overwrite them.")
            continue
        try:
            if use_db:
                plot_preprocess_tod_from_db(obsid, obs['group_by'], groups, configs, context, verbosity=verbosity)
            else:
                plot_preprocess_tod(obsid, configs, context, group_list=groups, verbosity=verbosity)
        except Exception as e:
            logger.info(f"{type(e)}: {e}")
            logger.info(''.join(traceback.format_tb(e.__traceback__)))
            logger.info(f'Skiping obs:{obs["obs_id"]} and moving to the next')
            continue

if __name__ == '__main__':
    sp_util.main_launcher(main, get_parser)



import numpy as np
import scipy.stats as stats
from scipy.signal import find_peaks

## "temporary" fix to deal with scipy>1.8 changing the sparse setup
try:
    from scipy.sparse import csr_array
except ImportError:
    from scipy.sparse import csr_matrix as csr_array

from so3g.proj import Ranges, RangesMatrix

from .. import core
from .. import coords
from . import filters
from . import fourier_filter 

def get_det_bias_flags(aman, detcal=None, rfrac_range=(0.1, 0.7),
                       psat_range=None, rn_range=None, si_nan=False,
                       merge=True, overwrite=True,
                       name='det_bias_flags', full_output=False):
    """
    Function for selecting detectors in appropriate bias range.

    Parameters
    ----------
    aman : AxisManager
        Input axis manager.
    detcal : AxisManager
        AxisManager containing detector calibration information
        from bias steps and IVs. If None defaults to aman.det_cal.
    rfrac_range : Tuple
        Tuple (lower_bound, upper_bound) for rfrac det selection.
    psat_range : Tuple
        Tuple (lower_bound, upper_bound) for P_SAT from IV analysis.
        P_SAT in the IV analysis is the bias power at 90% Rn in pW.
        If None, no flags are not applied from P_SAT. 
    rn_range : Tuple
        Tuple (lower_bound, upper_bound) for r_n det selection.
    si_nan : bool
        If true, flag dets where s_i is NaN. Default is false.
    merge : bool
        If true, merges the generated flag into aman.
    overwrite : bool
        If true, write over flag. If false, don't.
    name : str
        Name of flag to add to aman.flags if merge is True.
    full_output : bool
        If true, returns the full output with separated RangesMatrices

    Returns
    -------
    msk_aman : AxisManager
        AxisManager containing RangesMatrix shaped N_dets x N_samps
        that is True if the detector is flagged to be cut and false
        if it should be kept based on the rfrac, and psat ranges. 
        To create a boolean mask from the RangesMatrix that can be
        used for aman.restrict() use ``keep = ~has_all_cut(mask)``
        and then restrict with ``aman.restrict('dets', aman.dets.vals[keep])``.
        If full_output is True, this will contain multiple RangesMatrices.
    """
    if detcal is None:
        if 'det_cal' not in aman:
            raise ValueError("AxisManager missing required 'det_cal' field " 
                             "with detector calibration information")
        detcal = aman.det_cal

    if 'flags' not in aman:
        overwrite = False
        merge = False
    if overwrite and name in aman.flags:
        aman.flags.move(name, None)

    ranges = [detcal.bg >= 0,
              detcal.r_tes > 0,
              detcal.r_frac >= rfrac_range[0],
              detcal.r_frac <= rfrac_range[1]
             ]    
    if psat_range is not None:
        ranges.append(detcal.p_sat*1e12 >= psat_range[0])
        ranges.append(detcal.p_sat*1e12 <= psat_range[1])
    if rn_range is not None:
        ranges.append(detcal.r_n >= rn_range[0])
        ranges.append(detcal.r_n <= rn_range[1])
    if si_nan:
        ranges.append(np.isnan(detcal.s_i) == False)

    msk = ~(np.all(ranges, axis=0))
    # Expand mask to ndets x nsamps RangesMatrix
    if 'samps' in aman:
        x = Ranges(aman.samps.count)
        mskexp = RangesMatrix([Ranges.ones_like(x) if Y
                            else Ranges.zeros_like(x) for Y in msk])
        msk_aman = core.AxisManager(aman.dets, aman.samps)
        msk_aman.wrap(name, mskexp, [(0, 'dets'), (1, 'samps')])
    else:
        mskexp = msk
        msk_aman = core.AxisManager(aman.dets)
        msk_aman.wrap(name, mskexp, [(0, 'dets')])
    
    if merge:
        if name in aman.flags and not overwrite:
            raise ValueError(f"Flag name {name} already exists in aman.flags")
        if name in aman.flags:
            aman.flags[name] = mskexp
        else:
            aman.flags.wrap(name, mskexp, [(0, 'dets'), (1, 'samps')])

    if full_output:
        msks = []
        ranges = [detcal.bg >= 0,
                  detcal.r_tes > 0,
                  detcal.r_frac >= rfrac_range[0],
                  detcal.r_frac <= rfrac_range[1]
                 ]
        if psat_range is not None:
            ranges.append(detcal.p_sat*1e12 >= psat_range[0])
            ranges.append(detcal.p_sat*1e12 <= psat_range[1])
        
        if rn_range is not None:
            ranges.append(detcal.r_n >= rn_range[0])
            ranges.append(detcal.r_n <= rn_range[1])

        for range in ranges:
            msk = ~(np.all([range], axis=0))
            msks.append(RangesMatrix([Ranges.ones_like(x) if Y
                                      else Ranges.zeros_like(x) for Y in msk]))

        msk_names = ['bg', 'r_tes', 'r_frac_gt', 'r_frac_lt']
        
        if psat_range is not None:
            msk_names.extend(['p_sat_gt', 'p_sat_lt'])
            
        if rn_range is not None:
            msk_names.extend(['r_n_gt', 'r_n_lt'])

        for i, msk in enumerate(msks):
            if 'samps' in aman:
                msk_aman.wrap(f'{msk_names[i]}_flags', msk, [(0, 'dets'), (1, 'samps')])
            else:
                msk_aman.wrap(f'{msk_names[i]}_flags', msk, [(0, 'dets')])
    
    return msk_aman

def get_turnaround_flags(aman, az=None, method='scanspeed', name='turnarounds',
                         merge=True, merge_lr=True, overwrite=True, 
                         t_buffer=2., kernel_size=400, peak_threshold=0.1, rel_distance_peaks=0.3,
                         truncate=False, qlim=1, merge_subscans=True, turnarounds_in_subscan=False):
    """
    Compute turnaround flags for a dataset.

    Parameters
    ----------
    aman : AxisManager
        Input axis manager.
    az : Array
        (Optional). Azimuth data for turnaround flag computation. If not provided, it uses ``aman.boresight.az.``
    method : str
        (Optional). The method for computing turnaround flags. Options are ``az`` or ``scanspeed``.
    name : str
        (Optional). The name of the turnaround flag in ``aman.flags``. Default is ``turnarounds``
    merge : bool
        (Optional). Merge the computed turnaround flags into ``aman.flags`` if ``True``.
    merge_lr : bool
        (Optional). Merge left and right scan flags as ``aman.flags.left_scan`` and ``aman.flags.right_scan`` if ``True``.
    overwrite : bool
        (Optional). Overwrite an existing flag in ``aman.flags`` with the same name.
    t_buffer : float
        (Optional). Buffer time (in seconds) for flagging turnarounds in ``scanspeed`` method.
    kernel_size : int
        (Optional). Size of the step-wise matched filter kernel used in ``scanspeed`` method.
    peak_threshold : float
        (Optional). Peak threshold for identifying peaks in the matched filter response.
        It is a value used to determine the minimum peak height in the signal.
    rel_distance_peaks : float
        (Optional). Relative distance between peaks.
        It specifies the minimum distance between peaks as a fraction of the approximate number of samples in one scan period.
    truncate : bool
        (Optional). Truncate unstable scan segments if True in ``scanspeed`` method.
    qlim : float
        (Optional). Azimuth threshold percentile for ``az`` method turnaround detection.
    merge_subscans : bool
        (Optional). Also merge an AxisManager with subscan information.
    turnarounds_in_subscan : bool
        (Optional). Turnarounds are included as part of a subscan.

    Returns
    -------
    Ranges : RangesMatrix
        The turnaround flags as a Ranges object.
    """
    if az is None : 
        az = aman.boresight.az
        
    if method not in ['az', 'scanspeed']:
        raise ValueError('Unsupported method. Supported methods are `az` or `scanspeed`')
    
    # `az` method: flag turnarounds based on azimuth threshold specifled by qlim
    elif method=='az':
        lo, hi = np.percentile(az, [qlim, 100 - qlim])
        m = np.logical_or(az < lo, az > hi)
        ta_flag = Ranges.from_bitmask(m)
    
    # `scanspeed` method: flag turnarounds based on scanspeed.
    elif method=='scanspeed':
        daz = np.diff(az)
        daz = np.append(daz,daz[-1])
        approx_daz = np.median(daz[daz>np.percentile(daz, 95)])
        
        # derive approximate number of samples in one_scan period
        approx_samps_onescan = int(np.ptp(az) / approx_daz)
        # update approx_samps_onescan with detrending
        x = np.linspace(0, 1, az.shape[0])
        slope = az[-approx_samps_onescan:].mean() - az[:approx_samps_onescan].mean()
        approx_samps_onescan = int(np.ptp(az - slope*x) / approx_daz)

        # Make a step-function like matched filter. Kernel is normarized to make peak height ~1
        kernel = np.ones(kernel_size) / approx_daz / kernel_size
        kernel[kernel_size//2:] *= -1

        # convolve signal with the kernel
        pad_init = np.ones(kernel_size//2) * daz[0]
        pad_last = np.ones(kernel_size//2) * daz[-1]
        matched = np.convolve(np.hstack([pad_init, daz, pad_last]), kernel, mode='same')
        matched = matched[kernel_size//2:-kernel_size//2]

        # find peaks in matched daz
        peaks, _ = find_peaks(np.abs(matched), height=peak_threshold, distance=rel_distance_peaks*approx_samps_onescan)
        is_pos = matched[peaks] > 0

        # update approx_samps_onescan
        approx_samps_onescan = int(np.mean(np.diff(peaks)))
        
        # flags turnarounds, left/right scans
        _ta_flag = np.zeros(aman.samps.count, dtype=bool)
        _left_flag = np.zeros(aman.samps.count, dtype=bool)
        _right_flag = np.zeros(aman.samps.count, dtype=bool)
        
        dt = np.mean(np.diff(aman.timestamps))
        nbuffer_half = int(t_buffer/dt//2)
        
        for ip,p in enumerate(peaks[:-1]):
            _ta_flag[p-nbuffer_half:p+nbuffer_half] = True
            if is_pos[ip]: 
                _right_flag[peaks[ip]:peaks[ip+1]] = True
            else:
                _left_flag[peaks[ip]:peaks[ip+1]] = True
        _ta_flag[peaks[-1]-nbuffer_half:peaks[-1]+nbuffer_half] = True
        
        # Check the initial/last part. If the daz is the same as the other scaning part,
        # the part is regarded as left or right scan. If not, flagged as `_truncate_flag`, which
        # will be truncated if `truncate` is True, or flagged as `turnarounds` if `truncate` is False.
        _truncate_flag = np.zeros(aman.samps.count, dtype=bool)
        daz_right = daz[_right_flag & ~_ta_flag]
        daz_right_mean, daz_right_std, daz_right_samps = daz_right.mean(), daz_right.std(), daz_right.shape[0]
        daz_left = daz[_left_flag & ~_ta_flag]
        daz_left_mean, daz_left_std, daz_left_samps = daz_left.mean(), daz_left.std(), daz_left.shape[0]
        
        part_slices_ta_masked = [slice(None, np.where(_ta_flag)[0][0]), slice(np.where(_ta_flag)[0][-1], None)]
        part_slices_ta_unmasked = [slice(None, peaks[0]), slice(peaks[-1], None)]
        
        for part_slice_ta_masked, part_slice_ta_unmasked in zip(part_slices_ta_masked, part_slices_ta_unmasked):
            daz_part = daz[part_slice_ta_masked]
            daz_part_mean, daz_part_std, daz_part_samps = daz_part.mean(), daz_part.std(), daz_part.shape[0]
            if np.isclose(daz_part_mean, daz_right_mean, rtol=0.01, atol=3.*daz_right_std/np.sqrt(daz_right_samps)) and \
                np.isclose(daz_part_std, daz_right_std, rtol=3., atol=0):
                _right_flag[part_slice_ta_unmasked] = True
            elif np.isclose(daz_part_mean, daz_left_mean, rtol=0.01, atol=3.*daz_right_std/np.sqrt(daz_left_samps)) and \
                np.isclose(daz_part_std, daz_left_std, rtol=3., atol=0):
                _left_flag[part_slice_ta_unmasked] = True
            else:
                _truncate_flag[part_slice_ta_unmasked] = True
        
        # Check if flagging works
        check_sum = _left_flag.astype(int) + _right_flag.astype(int) + _truncate_flag.astype(int)
        check_sum = np.all(np.ones(aman.samps.count, dtype=int) == check_sum)
        if not check_sum:
            raise ValueError('Check sum failed. There are samples not allocated any of left, right, or truncate.')
        
        # merge left/right mask
        left_flag = Ranges.from_bitmask(_left_flag)
        right_flag = Ranges.from_bitmask(_right_flag)
        if merge_lr:
            if ("left_scan" in aman.flags or "right_scan" in aman.flags ) and not overwrite:
                raise ValueError("Flag name left/right_flag already exists in aman.flags.")
            else : 
                if "left_scan" in aman.flags:
                    aman.flags["left_scan"] = left_flag
                else :
                    aman.flags.wrap("left_scan", left_flag)
                if "right_scan" in aman.flags:
                    aman.flags["right_scan"] = right_flag
                else :
                    aman.flags.wrap("right_scan", right_flag)

        # truncate unstable scan before the first turnaround or after the last turnaround
        if truncate:
            valid_slice = slice(*np.where(~_truncate_flag)[0][[0, -1]])
            valid_i_start, valid_i_end = np.where(~_truncate_flag)[0][0], np.where(~_truncate_flag)[0][-1]
            aman.restrict('samps', (aman.samps.offset + valid_i_start, aman.samps.offset+valid_i_end))
            ta_flag = Ranges.from_bitmask(_ta_flag[valid_slice])
        else:
            ta_flag = Ranges.from_bitmask(np.logical_or(_ta_flag, _truncate_flag))
    
    # merge turnaround flags
    if merge:
        if name in aman.flags and not overwrite:
            raise ValueError("Flag name {} already exists in aman.flags".format(name))
        elif name in aman.flags:
            aman.flags[name] = ta_flag
        else:
            aman.flags.wrap(name, ta_flag)   

    if merge_subscans:
        get_subscans(aman, merge=True, include_turnarounds=turnarounds_in_subscan)

    if method == 'az':
        ta_exp = RangesMatrix([ta_flag for i in range(aman.dets.count)])
        return ta_exp
    if method == 'scanspeed':
        ta_exp = RangesMatrix([ta_flag for i in range(aman.dets.count)])
        left_exp = RangesMatrix([left_flag for i in range(aman.dets.count)])
        right_exp = RangesMatrix([right_flag for i in range(aman.dets.count)])
        return ta_exp, left_exp, right_exp
    
def get_glitch_flags(aman,
                     t_glitch=0.002,
                     hp_fc=0.5,
                     n_sig=10,
                     buffer=200,
                     detrend=None,
                     signal_name=None,
                     merge=True,
                     overwrite=False,
                     name="glitches",
                     full_output=False,
                     edge_guard=2000,
                     subscan=False):
    """
    Find glitches with fourier filtering. Translation from moby2 as starting point

    Parameters
    ----------
    aman : AxisManager
        The tod.
    t_glitch : float
        Gaussian filter width.
    hp_fc : float
        High pass filter cutoff.
    n_sig : int or float
        Significance of detection.
    buffer : int
        Amount to buffer flags around found location
    detrend : str
        Detrend method to pass to fourier_filter
    signal_name : str
        Field name in aman to detect glitches on if None, defaults to ``signal``
    merge : bool)
        If true, add to ``aman.flags``
    name : string
        Name of flag to add to ``aman.flags``
    overwrite : bool
        If true, write over flag. If false, raise ValueError if name already exists in AxisManager
    full_output : bool
        If true, return sparse matrix with the significance of the detected glitches
    edge_guard : int
        Number of samples at the beginning and end of the tod to exclude from
        the returned glitch RangesMatrix. Defaults to 2000 samples (10 sec).
    subscan : bool
        If True, compute the glitch threshold on a per-subscan basis. Includes turnarounds.

    Returns
    -------
    flag : RangesMatrix
        RangesMatrix object containing glitch mask.
    """

    if signal_name is None:
        signal_name = "signal"
    # f-space filtering
    filt = filters.high_pass_sine2(cutoff=hp_fc) * filters.gaussian_filter(t_sigma=t_glitch)
    fvec = fourier_filter(
        aman, filt, detrend=detrend, signal_name=signal_name, resize="zero_pad"
    )
    # get the threshods based on n_sig x nlev = n_sig x iqu x 0.741
    fvec = np.abs(fvec)
    if fvec.shape[-1] > 50000:
        ds = int(fvec.shape[-1]/20000)
    else: 
        ds = 1

    if subscan:
        # We include turnarounds
        subscan_indices = np.concatenate([aman.flags.left_scan.ranges(), (~aman.flags.left_scan).ranges()])
    else:
        subscan_indices = np.array([[0, fvec.shape[-1]]])

    msk = np.zeros_like(fvec, dtype='bool')
    iqr_ranges = np.zeros_like(fvec)
    if fvec.ndim == 1:
        for ss in subscan_indices:
            iqr_range = 0.741 * stats.iqr(fvec[ss[0]:ss[1]:ds])
            # get flags
            msk[ss[0]:ss[1]] = fvec[ss[0]:ss[1]] > iqr_range * n_sig
            iqr_ranges[ss[0]:ss[1]] = iqr_range
        msk[:edge_guard] = False
        msk[-edge_guard:] = False
        flag = Ranges.from_bitmask(msk)
    else:
        for ss in subscan_indices:
            iqr_range = 0.741 * stats.iqr(fvec[:,ss[0]:ss[1]:ds], axis=1)
            # get flags
            msk[:,ss[0]:ss[1]] = fvec[:,ss[0]:ss[1]] > iqr_range[:, None] * n_sig
            iqr_ranges[:,ss[0]:ss[1]] = iqr_range[:, None]
        msk[:,:edge_guard] = False
        msk[:,-edge_guard:] = False
        flag = RangesMatrix([Ranges.from_bitmask(m) for m in msk])
    flag.buffer(buffer)

    if merge:
        if name in aman.flags and not overwrite:
            raise ValueError("Flag name {} already exists in tod.flags".format(name))
        elif name in aman.flags:
            aman.flags[name] = flag
        else:
            aman.flags.wrap(name, flag)

    if full_output:
        if fvec.ndim == 1:
            # do not compress into csr_array
            glitches = core.AxisManager(aman.samps)
            data = np.zeros_like(fvec)
            data[msk] = fvec[msk] / iqr_ranges[msk]
            glitches.wrap("glitch_flags", flag, [(0, "samps")])
            glitches.wrap("glitch_detection", data, [(0, "samps")])
        else:
            indptr = np.append(
                0, np.cumsum([np.sum(msk[i]) for i in range(aman.dets.count)])
            )
            indices = np.concatenate([np.where(msk[i])[0] for i in range(aman.dets.count)])
            data = np.concatenate(
                [fvec[i][msk[i]] / iqr_ranges[i][msk[i]] for i in range(aman.dets.count)]
            )
            smat = csr_array(
                (data, indices, indptr), shape=(aman.dets.count, aman.samps.count)
            )
            glitches = core.AxisManager(
                aman.dets,
                aman.samps,
            )
            glitches.wrap("glitch_flags", flag, [(0, "dets"), (1, "samps")])
            glitches.wrap("glitch_detection", smat, [(0, "dets"), (1, "samps")])
        return flag, glitches

    return flag


def get_trending_flags(aman,
                       max_trend=1.2,
                       t_piece=500,
                       max_samples=500,
                       signal=None,
                       timestamps=None,
                       merge=True,
                       overwrite=True,
                       name="trends",
                       full_output=False):
    """
    Flag Detectors with trends larger than max_trend. 
    This function can be used to find unlocked detectors.
    Note that this is a rough cut and unflagged detectors can still have poor tracking.

    Parameters
    ----------
    aman : AxisManager
        The tod
    max_trend : float
        Slope at which detectors are unlocked. The default is for use with phase units.
    t_piece : float
        Duration in seconds of each pieces to cut the timestream in to to look for trends
    max_samples : int
        Maximum samples to compute the slope with.
    signal : array
        (Optional). Signal to use to generate flags, if None default is aman.signal.
    timestamps : array
        (Optional). Timestamps to use to generate flags, default is aman.timestamps.
    merge : bool
        If true, merges the generated flag into aman.
    overwrite : bool
        If true, write over flag. If false, don't.
    name : str
        Name of flag to add to aman.flags if merge is True.
    full_output : bool
        If true, returns calculated slope sizes

    Returns
    -------
    cut : RangesMatrix
        RangesMatrix of trending regions
    trends : AxisManager
        If full_output is true, calculated slopes and the sample edges where they were calculated.
    """
    if 'flags' not in aman:
        overwrite = False
        merge = False
    if overwrite and name in aman.flags:
        aman.flags.move(name, None)

    if signal is None:
        signal = aman.signal
    signal = np.atleast_2d(signal)
    if timestamps is None:
        timestamps = aman.timestamps
    assert len(timestamps) == signal.shape[1]
    
    # This helps with floating point precision
    # Not modifying inplace since we don't want to touch aman.timestamps
    timestamps = timestamps - timestamps[0]

    cut = np.zeros((len(signal), 0), dtype=bool)
    samp_edges = [0]
    
    # Get sampling rate
    fs = 1. / np.nanmedian(np.diff(timestamps))
    n_samples_per_piece = int(t_piece * fs)
    # How many pieces can timestamps be divided into
    n_pieces = len(timestamps) // n_samples_per_piece

    cut_list = []
    slope_list = []
    samp_edges = [0]

    split_ts = np.array_split(timestamps, n_pieces)
    split_sig = np.array_split(signal, n_pieces, axis=1)

    for t, s in zip(split_ts, split_sig):
        samps = len(t)
        if samps > max_samples:
            n = samps // max_samples
            t = t[::n]
            s = s[:, ::n]

        t_mean = t.mean()
        t_var = (t ** 2).mean() - t_mean ** 2
        s_mean = s.mean(axis=1)
        ts_mean = (t * s).mean(axis=1)

        _slopes = (ts_mean - t_mean * s_mean) / t_var
        slope_mask = np.abs(_slopes) > max_trend

        cut_chunk = np.broadcast_to(slope_mask[:, np.newaxis], (slope_mask.size, samps))
        cut_list.append(cut_chunk)

        if full_output:
            slope_list.append(_slopes)
            samp_edges.append(samp_edges[-1] + samps)

    cut = RangesMatrix.from_mask(np.hstack(cut_list))

    if merge:
        if name in aman.flags and not overwrite:
            raise ValueError("Flag name {} already exists in aman.flags".format(name))
        if name in aman.flags:
            aman.flags[name] = cut
        else:
            aman.flags.wrap(name, cut)

    if full_output:
        slopes = np.stack(slope_list, axis=1)
        samp_edges = np.array(samp_edges)

        trends = core.AxisManager(
            aman.dets,
            core.OffsetAxis("samps", len(timestamps)),
            core.OffsetAxis("trend_bins", len(samp_edges) - 1),
        )
        trends.wrap("samp_start", samp_edges[:-1], [(0, "trend_bins")])
        trends.wrap("trends", slopes, [(0, "dets"), (1, "trend_bins")])
        trends.wrap("trend_flags", cut, [(0, "dets"), (1, "samps")])
        return cut, trends

    return cut

def get_dark_dets(aman, merge=True, overwrite=True, dark_flags_name='darks'):
    """
    Identify and flag dark detectors in the given aman object.

    Parameters
    ----------
    aman : AxisManager
        The tod.
    merge : bool, optional
        If True, merge the dark detector flags into the aman.flags. Default is True.
    overwrite : bool, optional
        If True, overwrite existing flags with the same name. Default is True.
    dark_flags_name : str, optional
        The name to use for the dark detector flags in aman.flags. Default is 'darks'.
    
    Returns
    -------
    mskdarks: RangesMatrix
        A matrix of ranges indicating the dark detectors.

    Raises
    ------
    ValueError
        If merge is True and dark_flags_name already exists in aman.flags and overwrite is False.
    """
    darks = np.array(aman.det_info.wafer.type != 'OPTC')
    x = Ranges(aman.samps.count)
    mskdarks = RangesMatrix([Ranges.ones_like(x) if Y
                                else Ranges.zeros_like(x) for Y in darks])
    
    if merge:
        if dark_flags_name in aman.flags and not overwrite:
            raise ValueError(f"Flag name {dark_flags_name} already exists in aman.flags")
        if dark_flags_name in aman.flags:
            aman.flags[dark_flags_name] = mskdarks
        else:
            aman.flags.wrap(dark_flags_name, mskdarks, [(0, 'dets'), (1, 'samps')])

    return mskdarks

def get_source_flags(aman, merge=True, overwrite=True, source_flags_name=None,
                     mask=None, center_on=None, res=None, max_pix=None):

    if res:
        res = np.radians(res/60) # config input in arcminutes

    source_flags = coords.planets.compute_source_flags(tod=aman, mask=mask,
                                                       center_on=center_on,
                                                       res=res, max_pix=max_pix)

    if merge:
        if source_flags_name is None:
            source_flags_name = center_on
        if source_flags_name in aman.flags and not overwrite:
            raise ValueError(f"Flag name {source_flags_name} already exists in aman.flags")
        if source_flags_name in aman.flags:
            aman.flags[source_flags_name] = source_flags
        else:
            aman.flags.wrap(source_flags_name, source_flags, [(0, 'dets'), (1, 'samps')])

    return source_flags

def get_ptp_flags(aman, signal_name='signal', kurtosis_threshold=5,
                  merge=False, overwrite=False, ptp_flag_name='ptp_flag',
                  outlier_range=(0.5, 2.)):
    """
    Returns a ranges matrix that indicates if the peak-to-peak (ptp) of
    the tod is valid based on the kurtosis of the distribution of ptps. The
    threshold is set by ``kurtosis_threshold``.

    Parameters
    ----------
    aman : AxisManager
        The tod
    signal_name : str
        Signal to estimate flags off of. Default is ``signal``.
    kurtosis_threshold : float
        Maximum allowable kurtosis of the distribution of peak-to-peaks.
        Default is 5.
    merge : bool
        Merge RangesMatrix into ``aman.flags``. Default is False.
    overwrite : bool
        Whether to write over any existing data in ``aman.flags[ptp_flag_name]`` 
        if merge is True. Default is False.
    ptp_flag_name : str
        Field name used when merge is True. Default is ``ptp_flag``.
    outlier_range : tuple
        (lower, upper) bound of the initial cut before estimating the kurtosis.

    Returns
    -------
    mskptps : RangesMatrix
        RangesMatrix of detectors with acceptable peak-to-peaks. 
        All ones if the detector should be cut.

    """
    det_mask = np.full(aman.dets.count, True, dtype=bool)
    ptps_full = np.ptp(aman[signal_name], axis=1)
    ratio = ptps_full/np.median(ptps_full)
    outlier_mask = (ratio<outlier_range[0]) | (outlier_range[1]<ratio)
    det_mask[outlier_mask] = False
    if np.any(np.logical_not(np.isfinite(aman[signal_name][det_mask]))):
        raise ValueError(f"There is a nan in {signal_name} in aman {aman.obs_info['obs_id']} !!!")
    while True:
        if len(aman.dets.vals[det_mask]) > 0:
            ptps = np.ptp(aman[signal_name][det_mask], axis=1)
        else:
            break
        kurtosis_ptp = stats.kurtosis(ptps)
        if np.abs(kurtosis_ptp) < kurtosis_threshold:
            break
        else:
            max_is_bad_factor = np.max(ptps)/np.median(ptps)
            min_is_bad_factor = np.median(ptps)/np.min(ptps)
            if max_is_bad_factor > min_is_bad_factor:
                det_mask[ptps_full >= np.max(ptps)] = False
            else:
                det_mask[ptps_full <= np.min(ptps)] = False
    x = Ranges(aman.samps.count)
    mskptps = RangesMatrix([Ranges.zeros_like(x) if Y
                             else Ranges.ones_like(x) for Y in det_mask])
    if merge:
        if ptp_flag_name in aman.flags and not overwrite:
            raise ValueError(f"Flag name {ptp_flag_name} already exists in aman.flags")
        if ptp_flag_name in aman.flags:
            aman.flags[ptp_flag_name] = mskptps
        else:
            aman.flags.wrap(ptp_flag_name, mskptps, [(0, 'dets'), (1, 'samps')])

    return mskptps

def get_inv_var_flags(aman, signal_name='signal', nsigma=5,
                      merge=False, overwrite=False, inv_var_flag_name='inv_var_flag'):
    """
    Returns a ranges matrix that indicates if the inverse variance (inv_var) of
    the tod is greater than ``nsigma`` away from the median.
    
    Parameters
    ----------
    aman : AxisManager
        The tod
    signal_name : str
        Signal to estimate flags off of. Default is ``signal``.
    nsigma : float
        Maximum allowable deviation from the median inverse variance.
        Default is 5.
    merge : bool
        Merge RangesMatrix into ``aman.flags``. Default is False.
    overwrite : bool
        Whether to write over any existing data in ``aman.flags[inv_var_flag_name]`` 
        if merge is True. Default is False.
    inv_var_flag_name : str
        Field name used when merge is True. Default is ``inv_var_flag``.

    Returns
    -------
    mskptps : RangesMatrix
        RangesMatrix of detectors with acceptable inverse variance. 
        All ones if the detector should be cut.

    """
    ivar = 1.0/np.var(aman[signal_name], axis=-1)
    sigma = (np.percentile(ivar,84) - np.percentile(ivar, 16))/2
    det_mask = ivar > np.median(ivar) + nsigma*sigma
    x = Ranges(aman.samps.count)
    mskinvar = RangesMatrix([Ranges.ones_like(x) if Y
                             else Ranges.zeros_like(x) for Y in det_mask])
    if merge:
        if inv_var_flag_name in aman.flags and not overwrite:
            raise ValueError(f"Flag name {inv_var_flag_name} already exists in aman.flags")
        if inv_var_flag_name in aman.flags:
            aman.flags[inv_var_flag_name] = mskinvar
        else:
            aman.flags.wrap(inv_var_flag_name, mskinvar, [(0, 'dets'), (1, 'samps')])

    return mskinvar

def get_subscans(aman, flags=None, merge=True, include_turnarounds=False, overwrite=True):
    """
    Returns an axis manager with information about subscans.
    This includes direction and a ranges matrix (subscans samps)
    True inside each subscan.

    Parameters
    ----------
    aman : AxisManager
        Input AxisManager.
    flags : FlagManager
        AxisManager or FlagManager containing (samps,) Ranges left_scan, right_scan, and turnarounds. Defaults to aman.flags.
    merge : bool
        Merge into aman as 'subscan_info'
    include_turnarounds : bool
        Include turnarounds in the subscan ranges
    overwrite : bool
        If true, write over subscan_info.

    Returns
    -------
    subscan_aman : AxisManager
        AxisManager containing information about the subscans.
        "direction" is a (subscans,) array of strings 'left' or 'right'
        "subscan_flags" is a (subscans, samps) RangesMatrix; True inside the subscan.
    """
    if flags is None:
        flags = aman.flags
    if not include_turnarounds:
        ss_ind = (~flags.turnarounds).ranges() # sliceable indices (first inclusive, last exclusive) for subscans
    else:
        left = flags.left_scan.ranges()
        right = flags.right_scan.ranges()
        start_left = 0 if (left[0,0] < right[0,0]) else 1
        ss_ind = np.empty((left.shape[0] + right.shape[0], 2), dtype=left.dtype)
        ss_ind[start_left::2] = left
        ss_ind[(start_left-1)%2::2] = right

    start_inds, end_inds = ss_ind.T
    n_subscan = ss_ind.shape[0]
    subscan_aman = core.AxisManager(aman.samps, core.IndexAxis("subscans", n_subscan))

    is_left = flags.left_scan.mask()[start_inds]
    subscan_aman.wrap('direction', np.array(['left' if is_left[ii] else 'right' for ii in range(n_subscan)]), [(0, 'subscans')])

    rm = RangesMatrix([Ranges.from_array(np.atleast_2d(ss), aman.samps.count) for ss in ss_ind])
    subscan_aman.wrap('subscan_flags', rm, [(0, 'subscans'), (1, 'samps')]) # True in the subscan
    if merge:
        name = 'subscan_info'
        if overwrite and name in aman:
            aman.move(name, None)
        aman.wrap(name, subscan_aman)
    return subscan_aman

def get_subscan_signal(aman, arr, isub=None, trim=False):
    """
    Split an array into subscans.

    Parameters
    ----------
    aman : AxisManager
        Input AxisManager.
    arr : Array
        Input array.
    isub : int
        Index of the desired subscan. May also be a list of indices.
        If None, all are used.
    trim : bool
        Do not include size-zero arrays from empty subscans in the output.

    Returns
    -------
    out : list
        If isub is a scalar, return an Array of arr cut on the samps axis to the given subscan.
        If isub is a list or None, return a list of such Arrays.
    """
    if isinstance(arr, str):
        arr = aman[arr]
    if np.isscalar(isub):
        out = apply_rng(arr, aman.subscan_info.subscan_flags[isub])
        if trim and out.size == 0:
            out = None
    else:
        if isub is None:
            isub = range(len(aman.subscan_info.subscan_flags))
        out = [apply_rng(arr, aman.subscan_info.subscan_flags[ii]) for ii in isub]
        if trim:
            out = [x for x in out if x.size > 0]

    return out


def apply_rng(arr, rng):
    """
    Apply a Ranges object to an array. rng should be True on the samples you want to keep.

    Parameters
    ----------
    arr : Array
        Array containing the signal. Should have one axis of len (samps).
    rng : Ranges
        Ranges object of len (samps) selecting the desired range.
    """
    if rng.ranges().size == 0:
        slices = [slice(0,0)] #  Return an empty array if rng is empty
    else:
        slices = [slice(*irng) for irng in rng.ranges()]

    # Identify the samps axis
    isamps = np.where(np.array(arr.shape) == rng.count)[0]
    if isamps.size != 1:
        # Check for axis mismatch between arr and rng, or multiple axes with the same size
        raise RuntimeError("Could not identify axis matching Ranges")
    # Apply ranges
    out = []
    for slc in slices:
        ndslice = tuple((slice(None) if ii != isamps[0] else slc for ii in range(arr.ndim)))
        out.append(arr[ndslice])
    return np.concatenate(out, axis=isamps[0])

def wrap_stats(aman, info_aman_name, info, info_names, merge=True):
    """
    Wrap multiple stats into a new aman, checking for subscan information. Stats can be (dets,) or (dets, subscans).

    Parameters
    ----------
    aman : AxisManager
        Input AxisManager.
    info_aman_name : str
        Name for info_aman when wrapped into input.
    info : Array
        (stats, dets,) or (stats, dets, subscans) containing the information you want to wrap.
    info_names : list
        List of str names for each entry in the new aman.
    merge : bool
        If True merge info_aman into aman.

    Returns
    -------
    info_aman : AxisManager
        (dets,) or (dets, subscans) aman with a field for each item in info_names.
    """
    info_names = np.atleast_1d(info_names)
    info = np.atleast_2d(info)
    if info.shape == (len(info_names), aman.dets.count): # (stats, dets)
        if len(info_names) == aman.dets.count and aman.dets.count == aman.subscan_info.subscans.count:
            raise RuntimeError("Cannot infer axis mapping") # Catch corner case
        info_aman = core.AxisManager(aman.dets)
        axmap = [(0, 'dets')]

    else:
        info = np.atleast_3d(info) # (stats, dets, subscans)
        info_aman = core.AxisManager(aman.dets, aman.subscan_info.subscans)
        axmap = [(0, 'dets'), (1, 'subscans')]

    for ii in range(len(info_names)):
        info_aman.wrap(info_names[ii], info[ii], axmap)
    if merge:
        if info_aman_name in aman.keys():
            aman[info_aman_name].merge(info_aman)
        else:
            aman.wrap(info_aman_name, info_aman)
    return info_aman

def get_stats(aman, signal, stat_names, split_subscans=False, mask=None, name="stats", merge=False):
    """
    Calculate basic statistics on a TOD or power spectrum.

    The statistics currently implemented are:
    - ``'mean'``
    - ``'median'``
    - ``'ptp'`` (peak to peak)
    - ``'std'`` (standard deviation)
    - ``'var'`` (variance)
    - ``'kurtosis'``
    - ``'skew'``

    Parameters
    ----------
    aman : AxisManager
        Input AxisManager.
    signal : Array
        Input signal. Statistics will be computed over *axis 1*.
    stat_names : list
        List of strings identifying which statistics to run.
    split_subscans : bool
        If True statistics will be computed on subscans. Assumes aman.subscan_info exists already.
    mask : Array
        Mask to apply before computation. 1d array for advanced indexing (keep True), or a slice object.
    name : str
        Name of axis manager to add to aman if merge is True.
    """
    stat_names = np.atleast_1d(stat_names)
    fn_dict = {'mean': np.mean, 'median': np.median, 'ptp': np.ptp, 'std': np.std, 'var': np.var,
                     'kurtosis': stats.kurtosis, 'skew': stats.skew}

    if isinstance(signal, str):
        signal = aman[signal]
    if split_subscans:
        if mask is not None:
            raise ValueError("Cannot mask samples and split subscans")
        stats_arr = []
        for iss in range(aman.subscan_info.subscans.count):
            data = get_subscan_signal(aman, signal, iss)
            if data.size > 0:
                stats_arr.append([fn_dict[name](data, axis=1) for name in stat_names]) # Samps axis assumed to be 1
            else:
                stats_arr.append(np.full((len(stat_names), signal.shape[0]), np.nan)) # Add nans if subscan has been entirely cut
        stats_arr = np.array(stats_arr).transpose(1, 2, 0) # stat, dets, subscan
    else:
        if mask is None:
            mask = slice(None)
        stats_arr = np.array([fn_dict[name](signal[:, mask], axis=1) for name in stat_names]) # Samps axis assumed to be 1

    info_aman = wrap_stats(aman, name, stats_arr, stat_names, merge)
    return info_aman

def get_focalplane_flags(aman, merge=True, overwrite=True, invalid_flags_name='fp_flags'):
    """
    Generate flags for invalid detectors in the focal plane.
        The tod.
    merge : bool
        If true, merges the generated flag into aman.
    overwrite : bool
        If true, write over flag. If false, don't.
    invalid_flags_name : str
        Name of flag to add to aman.flags if merge is True.

    Returns
    -------
    msk_invalid_fp : RangesMatrix
        RangesMatrix of invalid detectors in the focal plane.
    """
    # Available detectors in focalplane
    xi_nan = np.isnan(aman.focal_plane.xi)
    eta_nan = np.isnan(aman.focal_plane.eta)
    gamma_nan = np.isnan(aman.focal_plane.gamma)
    x = Ranges(aman.samps.count)
    flag_invalid_fp = np.sum([xi_nan, eta_nan, gamma_nan], axis=0) != 0
    msk_invalid_fp = RangesMatrix([Ranges.ones_like(x) if Y else Ranges.zeros_like(x) for Y in flag_invalid_fp])
    
    if merge:
        if invalid_flags_name in aman.flags and not overwrite:
            raise ValueError(f"Flag name {invalid_flags_name} already exists in aman.flags")
        if invalid_flags_name in aman.flags:
            aman.flags[invalid_flags_name] = msk_invalid_fp
        else:
            aman.flags.wrap(invalid_flags_name, msk_invalid_fp, [(0, 'dets'), (1, 'samps')])

    return msk_invalid_fp

def noise_fit_flags(aman, low_wn, high_wn, high_fk):
    """
    Evaluate white noise and fknee cuts based on provided boundaries.

    Parameters:
    aman : object
        An object containing noise fit statistics and noise model coefficients.
    low_wn : float or None
        The lower boundary for white noise. If None, white noise flagging is skipped.
    high_wn : float or None
        The upper boundary for white noise. If None, white noise flagging is skipped.
    high_fk : float or None
        The upper boundary for fknee. If None, fknee flagging is skipped.

    Returns:
    tuple or None
        A tuple containing flags for valid white noise and fknee if both boundaries are provided.
        If only one boundary is provided, returns the corresponding flag.
        If no boundaries are provided, returns None.
    """
    noise = aman.noise_fit_stats_signal.fit
    fk = noise[:, 0]
    wn = noise[:, 1]
    if low_wn is None:
        print(f"white noise boundaries are not defined, skipping.")
        flag_valid_wn = None
    else:
        flag_valid_wn = (low_wn < wn * 1e6) & (wn * 1e6 < high_wn)
    if high_fk is None:
        print(f"fknee boundaries are not defined, skipping.")
        flag_valid_fk = None
    else:
        flag_valid_fk = fk < high_fk
    if low_wn is not None and high_fk is not None:
        return flag_valid_wn, flag_valid_fk
    elif low_wn is not None:
        return flag_valid_wn
    elif high_fk is not None:
        return flag_valid_fk
    else:
        return None

def get_noisy_subscan_flags(aman, subscan_stats, nstd_lim=None,
                         ptp_lim=None, kurt_lim=None,
                         skew_lim=None, noisy_detector_lim=0.9,
                         merge=False, overwrite=False, 
                         name="noisy_subscan"):
    """
    Identify and flag bad subscans based on various statistical thresholds.
    aman : AxisManager 
        The tod.
    subscan_stats : dict
        Dictionary containing statistical metrics for subscans. Keys should 
        include 'std', 'ptp', 'kurtosis', and 'skew'.
    nstd_lim : float [optional]
        Threshold for standard deviation.
    ptp_lim : float [optional]
        Threshold for peak-to-peak values in pW.
    kurt_lim : float [optional]
        Threshold for kurtosis.
    skew_lim : float [optional]
        Threshold for skewness.
    noisy_detector_lim : float [optional]
        If > noisy_detector_lim fraction of samps are in flagged subscans
        then the detector is added to noisy_detector_flags.
    merge : bool
        If true, merges the generated flag into aman.
    overwrite : bool
        If true, write over flag. If false, don't.
    name : str
        Name of flag to add to aman.flags if merge is True.

    Returns
    -------
    noisy_subscan_flags : RangesMatrix
        RangesMatrix of bad subscans. (dets, samps), True for flagged (bad) samples.
    noisy_detector_flags : ndarray
        Array indicating detectors that are too noisy for more than noisy_detector_lim fraction
        of the obs duration. (dets,). We return ~noisy_detector_flags which is True for good detectors.

    """

    if 'flags' not in aman:
        overwrite = False
        merge = False
    if overwrite and name in aman.flags:
        aman.flags.move(name, None)
    
    median_std = np.median(subscan_stats["std"], axis=1)[:, np.newaxis]
    noisy_subscan_indicator = np.zeros_like(subscan_stats["std"], dtype=bool)
    
    if ptp_lim is not None and 'ptp' in subscan_stats:
        noisy_subscan_indicator |= (subscan_stats['ptp'] > ptp_lim)
    
    if nstd_lim is not None and 'std' in subscan_stats:
        noisy_subscan_indicator |= (subscan_stats['std'] > median_std * nstd_lim)
    
    if kurt_lim is not None and 'kurtosis' in subscan_stats:
        noisy_subscan_indicator |= (np.abs(subscan_stats['kurtosis']) > kurt_lim)
    
    if skew_lim is not None and 'skew' in subscan_stats:
        noisy_subscan_indicator |= (np.abs(subscan_stats['skew']) > skew_lim)

    if "subscan_info" in aman:
        ssf = aman.subscan_info.subscan_flags
    else:
        subscan_aman = get_subscans(aman, merge=False)
        ssf = subscan_aman.subscan_flags

    zeros=RangesMatrix.zeros((1, aman.samps.count)) # Needed when no subscans are True
    # For each det, select flagged subscans and merge their Ranges
    noisy_subscan_flags = RangesMatrix([np.sum(RangesMatrix.concatenate([ssf[noisy_subscan_indicator[idet]], zeros]), axis=0) for idet in range(aman.dets.count)])
    # Detectors which are too noisy for > noisy_detector_lim fraction of the obs duration
    noisy_detector_flags = np.array([np.sum(np.diff(rng.ranges(), axis=1)) / aman.samps.count for rng in noisy_subscan_flags.ranges]) > noisy_detector_lim
    
    if merge:
        if name in aman.flags and not overwrite:
            raise ValueError(f"Flag name {name} already exists in aman.flags")
        aman.flags.wrap(name, noisy_subscan_flags)

    return noisy_subscan_flags, ~noisy_detector_flags


def expand_smurfgaps_flags(aman, buffer=200, name='smurfgaps', merge=True):
    """
    smurfgaps flags indicates the samples of each stream_id where the
    lost frames are filled in the bookbinding process.
    See `sotodlib.io.bookbinder.bind`.

    This function expands smurfgaps flags of each stream_id to all detectors.

    Parameters
    ----------
    aman: AxisManager
        Input AxisManager
    buffer: int
        Amount of buffer to apply on smurfgaps
    name: str
        Name of flag to add to aman.flags if merge is True.
    merge: bool
        If true, merges the generated flag into aman.

    Returns
    -------
    smurfgaps: RangesMatrix
        smurfgaps flag with 'dets' and 'samps' axis

    """

    smurfgaps = RangesMatrix([aman.flags.get('smurfgaps_' + ufm) for ufm in aman.det_info.stream_id])
    if buffer:
        smurfgaps = smurfgaps.buffer(buffer)
    if merge:
        aman.flags.wrap('smurfgaps', smurfgaps, [(0, 'dets'), (1, 'samps')])
    return smurfgaps
